# Copyright (c) 2023, TU Wien, Department of Geodesy and Geoinformation.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Vienna University of Technology,
#     Department of Geodesy and Geoinformation nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Classes for reading and writing point data in NetCDF files
using Climate Forecast Metadata Conventions (http://cfconventions.org/).
"""

import os

import numpy as np
import datetime
import netCDF4

from pygeobase.io_base import GriddedBase


class PointData:
    """
    PointData class for reading and writing netCDF files following the
    CF conventions for point data.

    Parameters
    ----------
    filename : str
        Filename of netCDF file. If already exiting then it will be opened
        as read only unless the append keyword is set.
    mode : str, optional
        access mode. default "r"
        "r" means read-only; no data can be modified.
        "w" means write; a new file is created, an existing file with the
            same name is deleted.
        "a" and "r+" mean append (in analogy with serial files); an existing
            file is opened for reading and writing.
        Appending s to modes w, r+ or a will enable unbuffered shared access
        to NETCDF3_CLASSIC or NETCDF3_64BIT formatted files. Unbuffered
        access may be useful even if you don"t need shared access, since it
        may be faster for programs that don"t access data sequentially.
        This option is ignored for NETCDF4 and NETCDF4_CLASSIC
        formatted files.
    zlib : boolean, optional
        If set netCDF compression will be used. Default True
    complevel : int, optional
        Compression level used from 1(low compression) to 9(high compression).
        Default: 4
    n_obs : int, optional
        Number of observations. If None, unlimited dimension will be used.
        Default: None
    obs_dim : str, optional
        Observation dimension name. Default: "obs"
    add_dims : dict, optional
        Additional dimensions. Default: None
    loc_id_var : str, optional
        Location id variable name. Default: "location id"
    time_units : str, optional
        Time unit.
    time_var : str, optional
        Time variable name. Default "time"
    lat_var : str, optional
        Latitude variable name. Default "lat"
    lon_var : str, optional
        Longitude variable name. Default: "lon"
    alt_var : str, optional
        Altitude variable name. Default: "alt"
    """

    def __init__(self,
                 filename,
                 mode="r",
                 file_format="NETCDF4",
                 zlib=True,
                 complevel=4,
                 n_obs=None,
                 obs_dim="obs",
                 add_dims=None,
                 loc_id_var="location_id",
                 time_units="days since 1900-01-01 00:00:00",
                 time_var="time",
                 lat_var="lat",
                 lon_var="lon",
                 alt_var="alt",
                 **kwargs):

        self.nc_finfo = {
            "filename": filename,
            "mode": mode,
            "format": file_format
        }

        self.nc = None

        initial_mode = mode

        if mode == "a" and not os.path.exists(filename):
            initial_mode = "w"

        if initial_mode == "w":
            path = os.path.dirname(filename)
            if not os.path.exists(path):
                os.makedirs(path)
        self.compression_info = {"zlib": zlib, "complevel": complevel}

        try:
            self.nc = netCDF4.Dataset(filename,
                                      format=file_format,
                                      mode=initial_mode)
        except RuntimeError:
            raise IOError(f"File {filename} does not exist.")

        loc_id_attr = {"long_name": "location_id"}

        lon_attr = {
            "standard_name": "longitude",
            "long_name": "location longitude",
            "units": "degrees_east",
            "valid_range": (-180.0, 180.0)
        }

        lat_attr = {
            "standard_name": "latitude",
            "long_name": "location latitude",
            "units": "degrees_north",
            "valid_range": (-90.0, 90.0)
        }

        alt_attr = {
            "standard_name": "height",
            "long_name": "vertical distance above the "
            "surface",
            "units": "m",
            "positive": "up",
            "axis": "Z"
        }

        time_attr = {"standard_name": "time"}

        self.obs_dim = obs_dim

        if add_dims is not None:
            self.dim = add_dims.copy()
            self.dim.update({obs_dim: n_obs})
        else:
            self.dim = {obs_dim: n_obs}

        self.var = {
            "loc_id": {
                "name": loc_id_var,
                "dim": obs_dim,
                "attr": loc_id_attr,
                "dtype": np.int32
            },
            "lon": {
                "name": lon_var,
                "dim": obs_dim,
                "attr": lon_attr,
                "dtype": np.float32
            },
            "lat": {
                "name": lat_var,
                "dim": obs_dim,
                "attr": lat_attr,
                "dtype": np.float32
            },
            "alt": {
                "name": alt_var,
                "dim": obs_dim,
                "attr": alt_attr,
                "dtype": np.float32
            },
            "time": {
                "name": time_var,
                "dim": obs_dim,
                "unit": time_units,
                "dtype": np.float64,
                "attr": time_attr
            }
        }

        self.builtin_vars = [self.var[key]["name"] for key in self.var]

        if initial_mode == "w":

            s = "%Y-%m-%d %H:%M:%S"
            attr = {
                "id": os.path.split(self.nc_finfo["filename"])[1],
                "date_created": datetime.datetime.now().strftime(s),
                "featureType": "point"
            }

            self.nc.setncatts(attr)
            self._create_dims(self.dim)
            self._init_loc_var()

        # find next free position, i.e. next empty loc_id
        self.loc_idx = 0
        if initial_mode in ["r+", "a"]:
            loc_id = self.nc.variables[self.var["loc_id"]["name"]]
            if self.nc.dimensions[obs_dim].isunlimited():
                self.loc_idx = loc_id.shape[0]
            else:
                self.loc_idx = np.where(loc_id[:].mask)[0][0]

    def __str__(self):
        """
        String representation of class instance.
        """
        if self.nc is not None:
            str = self.nc.__str__()
        else:
            str = "NetCDF file closed."

        return str

    def flush(self):
        """
        Flush data.
        """
        if self.nc is not None:
            if self.nc_finfo["mode"] in ["w", "r+", "a"]:
                self.nc.sync()

    def close(self):
        """
        Close file.
        """
        if self.nc is not None:
            self.flush()
            self.nc.close()
            self.nc = None

    def __enter__(self):
        """
        ContextManager enter.
        """
        return self

    def __exit__(self, value_type, value, traceback):
        """
        ContextManager exit.
        """
        self.close()

    def __del__(self):
        """
        Destructor.
        """
        self.close()

    def _create_dims(self, dims):
        """
        Create dimensions in NetCDF file.

        Parameters
        ----------
        dims : dict
            NetCDF dimension.
        """
        for name, size in dims.items():
            self.nc.createDimension(name, size=size)

    def _init_loc_var(self):
        """
        Initialize location information (lon, lat, etc.).
        """
        for k, var in self.var.items():
            self.nc.createVariable(var["name"],
                                   var["dtype"],
                                   dimensions=var["dim"],
                                   **self.compression_info)
            self.nc.variables[var["name"]].setncatts(var["attr"])

    def write(self,
              loc_id,
              data,
              lon=None,
              lat=None,
              alt=None,
              time=None,
              **kwargs):
        """
        Write data for specified location ids.

        Parameters
        ----------
        loc_id : numpy.ndarray
            Location id.
        data : dict of numpy.ndarray or numpy.recarray
            Dictionary containing variable names as keys and data as items.
        lon : numpy.ndarray, optional
            Longitude information. Default: None
        lat : numpy.ndarray, optional
            Latitude information. Default: None
        alt : numpy.ndarray, optional
            Altitude information. Default: None
        time : numpy.ndarray, optional
            Time information. Default: None
        """
        if self.nc_finfo["mode"] in ["w", "r+", "a"]:

            num = np.array(loc_id).size
            idx = slice(self.loc_idx, self.loc_idx + num)

            # convert dict to recarray
            if isinstance(data, dict):

                # collect metadata info
                sub_md_list = [v.dtype.metadata for v in data.values()]

                # collect dtype info
                dtype_list = [(str(k), data[k].dtype.str, data[k].shape)
                              for k in data.keys()]

                # merge metadata info into common dict
                md_dict = {}
                for md in sub_md_list:
                    if md is not None and "dims" in md:
                        md_dict.update(md["dims"])

                # convert dict to recarray
                metadata = {"dims": md_dict}
                dtype = np.dtype(dtype_list, metadata=metadata)
                data = np.rec.fromarrays(data.values(), dtype=dtype)

            for var_data in data.dtype.names:
                if var_data not in self.nc.variables:
                    dtype = data[var_data].dtype
                    dimensions = (self.obs_dim, )

                    # check if custom metadata is included
                    if data.dtype.metadata is not None:
                        metadata = data.dtype.metadata
                        if "dims" in metadata and var_data in metadata["dims"]:
                            dimensions = metadata["dims"][var_data]

                    comp = self.compression_info.copy()

                    if not np.issubdtype(dtype, np.number):
                        comp['zlib'] = False

                    self.nc.createVariable(var_data,
                                           dtype,
                                           dimensions=dimensions,
                                           **comp)

                self.nc.variables[var_data][idx] = data[var_data]

            var_loc_id = self.var["loc_id"]["name"]
            self.nc.variables[var_loc_id][idx] = loc_id

            if lon is not None:
                var_lon = self.var["lon"]["name"]
                self.nc.variables[var_lon][idx] = lon

            if lat is not None:
                var_lat = self.var["lat"]["name"]
                self.nc.variables[var_lat][idx] = lat

            if alt is not None:
                var_alt = self.var["alt"]["name"]
                self.nc.variables[var_alt][idx] = alt

            if time is not None:
                var_time = self.var["time"]["name"]
                self.nc.variables[var_time][idx] = time

            self.loc_idx += num
        else:
            raise IOError("Write operations failed. File not open for writing.")

    def read(self, loc_id):
        """
        Read variable from netCDF file for given location id.

        Parameters
        ----------
        loc_id : int
            Location id.

        Returns
        -------
        data : dict
            Dictionary containing variable names as a key and data as items.
        """
        data = None

        if self.nc_finfo["mode"] in ["r", "r+", "a"]:
            loc_id_var = self.nc.variables[self.var["loc_id"]["name"]][:]
            pos = np.where(loc_id_var == loc_id)[0]

            if pos.size > 0:
                data = {}
                for var_name in self.nc.variables.keys():
                    read_data = self.nc.variables[var_name][pos]
                    if var_name not in self.builtin_vars:
                        read_data = np.squeeze(read_data)

                    data[var_name] = read_data

        else:
            raise IOError("Read operations failed. File not open for reading.")

        return data

    def __getitem__(self, item):
        """
        Accessing netCDF variable.

        Parameters
        ----------
        item : str
            Variable name.

        Returns
        -------
        var : netcdf4.variable
            NetCDF variable.
        """
        return self.nc.variables[item]


class GriddedPointData(GriddedBase):
    """
    GriddedPointData class using GriddedBase class as parent and
    PointData as i/o class.
    """

    def __init__(self, *args, **kwargs):
        kwargs["ioclass"] = PointData
        if "fn_format" not in kwargs:
            kwargs["fn_format"] = "{:04d}.nc"
        super(GriddedPointData, self).__init__(*args, **kwargs)

    def to_point_data(self, filename, **kwargs):
        """
        Re-write gridded point data into single file.

        Parameters
        ----------
        filename : str
            File name.
        """
        with PointData(filename, mode="w", **kwargs) as nc:
            for data, gp in self.iter_gp():
                nc.write(gp,
                         data,
                         lon=data["lon"],
                         lat=data["lat"],
                         alt=data["alt"],
                         time=data["time"])



# Copyright (c) 2023, TU Wien, Department of Geodesy and Geoinformation.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Vienna University of Technology,
#     Department of Geodesy and Geoinformation nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Testing time series class.
"""

import os
import pytest
import unittest
import warnings
from datetime import datetime, timedelta
from tempfile import mkdtemp

import numpy as np
import numpy.testing as nptest
import pandas as pd
import pygeogrids.grids as grids

from pynetcf.time_series import OrthoMultiTs
from pynetcf.time_series import ContiguousRaggedTs
from pynetcf.time_series import IndexedRaggedTs
from pynetcf.time_series import GriddedNcOrthoMultiTs
from pynetcf.time_series import GriddedNcContiguousRaggedTs
from pynetcf.time_series import GriddedNcIndexedRaggedTs


def test_open_non_existing_file():
    classes = [OrthoMultiTs, ContiguousRaggedTs, IndexedRaggedTs]
    for cls in classes:
        with pytest.raises(IOError):
            cls("not_existing_fname.nc")


class OrthoMultiTest(unittest.TestCase):

    def setUp(self):
        self.testfilename = os.path.join(mkdtemp(), "test.nc")


#    def tearDown(self):
#        os.remove(self.testfilename)

    def test_file_io_simple(self):

        tunits = 'days since 2000-01-01 00:00:00'

        with OrthoMultiTs(self.testfilename, mode="w", n_loc=3,
                          time_units=tunits) as dataset:
            for n_data in [5]:
                for location in [1, 2, 3]:

                    data = {"test": np.arange(n_data)}
                    base = datetime(2007, 1, n_data)
                    dates = np.array(
                        [base + timedelta(hours=i) for i in range(n_data)])
                    dataset.write(location,
                                  data,
                                  dates,
                                  loc_descr="first station",
                                  lon=0,
                                  lat=0,
                                  alt=5)

        with OrthoMultiTs(self.testfilename) as dataset:
            assert dataset.time_units == tunits
            data = dataset.read_all(2)
            nptest.assert_array_equal(data["test"], np.arange(5))

            test_dates = []
            for n_data in [5]:
                base = datetime(2007, 1, n_data)
                test_dates.append(
                    np.array(
                        [base + timedelta(hours=i) for i in range(n_data)]))
            dates = np.concatenate(test_dates)
            nptest.assert_array_equal(data["time"], dates)

    def test_file_io_reading_non_existing_point(self):

        with OrthoMultiTs(self.testfilename, mode="w", n_loc=3) as dataset:
            for n_data in [5]:
                for location in [1, 2, 6]:

                    data = {"test": np.arange(n_data)}
                    base = datetime(2007, 1, n_data)
                    dates = np.array(
                        [base + timedelta(hours=i) for i in range(n_data)])
                    dataset.write(location,
                                  data,
                                  dates,
                                  loc_descr="first station",
                                  lon=0,
                                  lat=0,
                                  alt=5)

        with OrthoMultiTs(self.testfilename) as dataset:
            assert dataset.time_units == "days since 1900-01-01 00:00:00"
            with pytest.raises(IOError):
                data = dataset.read_all(5)

    def test_file_io_2_steps(self):

        with OrthoMultiTs(self.testfilename, n_loc=3, mode="w") as dataset:
            for n_data in [5]:
                for location in [1, 2, 3]:

                    data = {"test": np.arange(n_data)}
                    base = datetime(2007, 1, n_data)
                    dates = np.array(
                        [base + timedelta(hours=i) for i in range(n_data)])

                    dataset.write(location,
                                  data,
                                  dates,
                                  loc_descr="first station",
                                  lon=0,
                                  lat=0,
                                  alt=5,
                                  fill_values={"test": -1})

        with OrthoMultiTs(self.testfilename, n_loc=3, mode="a") as dataset:
            for n_data in [5]:
                for location in [1, 2, 3]:

                    data = {"test": np.arange(n_data) + n_data}
                    base = datetime(2007, 2, n_data)
                    dates = np.array(
                        [base + timedelta(hours=i) for i in range(n_data)])
                    dataset.write(location,
                                  data,
                                  dates,
                                  loc_descr="first station",
                                  lon=0,
                                  lat=0,
                                  alt=5,
                                  fill_values={"test": -1})

        with OrthoMultiTs(self.testfilename) as dataset:
            data = dataset.read_all(2)
            nptest.assert_array_equal(data["test"], np.arange(10))

            test_dates = []
            for n_data in [5]:
                base = datetime(2007, 1, n_data)
                test_dates.append(
                    np.array(
                        [base + timedelta(hours=i) for i in range(n_data)]))
                base = datetime(2007, 2, n_data)
                test_dates.append(
                    np.array(
                        [base + timedelta(hours=i) for i in range(n_data)]))
            dates = np.concatenate(test_dates)
            nptest.assert_array_equal(data["time"], dates)

    def test_file_write_all(self):

        with OrthoMultiTs(self.testfilename, n_loc=3, mode="w") as dataset:
            n_data = 5
            locations = np.array([1, 2, 3])
            data = {"test": np.arange(n_data * 3).reshape(3, n_data)}
            base = datetime(2007, 1, n_data)
            dates = np.array(
                [base + timedelta(hours=i) for i in range(n_data)])
            descriptions = np.repeat([str("station")], 3).tolist()

            dataset.write_all(locations,
                              data,
                              dates,
                              loc_descrs=descriptions,
                              lons=np.arange(3),
                              lats=np.arange(3),
                              alts=np.arange(3))

        with OrthoMultiTs(self.testfilename) as dataset:
            data = dataset.read_all(2)
            nptest.assert_array_equal(data["test"], np.arange(5) + 5)
            test_dates = []
            for n_data in [5]:
                base = datetime(2007, 1, n_data)
                test_dates.append(
                    np.array(
                        [base + timedelta(hours=i) for i in range(n_data)]))
            dates = np.concatenate(test_dates)
            nptest.assert_array_equal(data["time"], dates)

    def test_file_write_all_1_location(self):

        with OrthoMultiTs(self.testfilename, n_loc=1, mode="w") as dataset:
            n_data = 5
            locations = np.array([1])
            data = {"test": np.arange(n_data).reshape(1, n_data)}
            base = datetime(2007, 1, n_data)
            dates = np.array(
                [base + timedelta(hours=i) for i in range(n_data)])
            descriptions = np.repeat([str("station")], 1).tolist()

            dataset.write_all(locations,
                              data,
                              dates,
                              loc_descrs=descriptions,
                              lons=np.arange(1),
                              lats=np.arange(1),
                              alts=None)

        with OrthoMultiTs(self.testfilename) as dataset:
            data = dataset.read_all(1)
            nptest.assert_array_equal(data["test"], np.arange(5))
            test_dates = []
            for n_data in [5]:
                base = datetime(2007, 1, n_data)
                test_dates.append(
                    np.array(
                        [base + timedelta(hours=i) for i in range(n_data)]))
            dates = np.concatenate(test_dates)
            nptest.assert_array_equal(data["time"], dates)

    def test_file_write_all_attributes(self):

        with OrthoMultiTs(self.testfilename, n_loc=3, mode="w") as dataset:
            n_data = 5
            locations = np.array([1, 2, 3])
            data = {
                "test": np.arange(n_data * 3).reshape(3, n_data),
                "test2": np.arange(n_data * 3).reshape(3, n_data)
            }
            base = datetime(2007, 1, n_data)
            dates = np.array(
                [base + timedelta(hours=i) for i in range(n_data)])
            descriptions = np.repeat([str("station")], 3).tolist()

            dataset.write_all(locations,
                              data,
                              dates,
                              loc_descrs=descriptions,
                              lons=np.arange(3),
                              lats=np.arange(3),
                              alts=np.arange(3),
                              attributes={"testattribute": "teststring"})

        with OrthoMultiTs(self.testfilename) as dataset:
            data = dataset.read_all(2)
            nptest.assert_array_equal(data["test"], np.arange(5) + 5)
            assert dataset.dataset.variables[
                "test"].testattribute == "teststring"
            assert dataset.dataset.variables[
                "test2"].testattribute == "teststring"
            test_dates = []
            for n_data in [5]:
                base = datetime(2007, 1, n_data)
                test_dates.append(
                    np.array(
                        [base + timedelta(hours=i) for i in range(n_data)]))
            dates = np.concatenate(test_dates)
            nptest.assert_array_equal(data["time"], dates)

    def test_file_write_attributes_for_each(self):
        """
        test writing two datasets with attributes for each dataset
        """

        with OrthoMultiTs(self.testfilename, n_loc=3, mode="w") as dataset:
            n_data = 5
            locations = np.array([1, 2, 3])
            data = {
                "test": np.arange(n_data * 3).reshape(3, n_data),
                "test2": np.arange(n_data * 3).reshape(3, n_data)
            }
            base = datetime(2007, 1, n_data)
            dates = np.array(
                [base + timedelta(hours=i) for i in range(n_data)])
            descriptions = np.repeat([str("station")], 3).tolist()

            dataset.write_all(locations,
                              data,
                              dates,
                              loc_descrs=descriptions,
                              lons=np.arange(3),
                              lats=np.arange(3),
                              alts=np.arange(3),
                              attributes={
                                  "test": {
                                      "testattribute": "teststring"
                                  },
                                  "test2": {
                                      "testattribute2": "teststring2"
                                  }
                              })

        with OrthoMultiTs(self.testfilename) as dataset:
            data = dataset.read_all(2)
            nptest.assert_array_equal(data["test"], np.arange(5) + 5)
            assert dataset.dataset.variables[
                "test"].testattribute == "teststring"
            assert dataset.dataset.variables[
                "test2"].testattribute2 == "teststring2"
            test_dates = []
            for n_data in [5]:
                base = datetime(2007, 1, n_data)
                test_dates.append(
                    np.array(
                        [base + timedelta(hours=i) for i in range(n_data)]))
            dates = np.concatenate(test_dates)
            nptest.assert_array_equal(data["time"], dates)


class DatasetContiguousTest(unittest.TestCase):

    def setUp(self):
        self.testfilename = os.path.join(mkdtemp(), "test.nc")

    def tearDown(self):
        os.remove(self.testfilename)

    def test_file_writing(self):

        dates = np.array(
            [datetime(2007, 1, 1),
             datetime(2007, 2, 1),
             datetime(2007, 3, 1)])

        tunits = 'days since 2000-01-01 00:00:00'

        with ContiguousRaggedTs(self.testfilename, n_loc=3, n_obs=9, mode="w",
                                time_units=tunits) as dataset:
            data = {"test": np.arange(3)}
            dataset.write(1,
                          data,
                          dates,
                          loc_descr="first station",
                          lon=1,
                          lat=1,
                          alt=1)
            dataset.write(2,
                          data,
                          dates,
                          loc_descr="second station",
                          lon=2,
                          lat=2,
                          alt=2)
            dataset.write(3,
                          data,
                          dates,
                          loc_descr="third station",
                          lon=3,
                          lat=3,
                          alt=3)

        with ContiguousRaggedTs(self.testfilename) as dataset:
            assert dataset.time_units == tunits
            data = dataset.read_all(1)
            nptest.assert_array_equal(data["test"], np.arange(3))
            nptest.assert_array_equal(data["time"], dates)

    def test_unlim_obs_file_writing(self):

        dates = np.array(
            [datetime(2007, 1, 1),
             datetime(2007, 2, 1),
             datetime(2007, 3, 1)])

        with ContiguousRaggedTs(self.testfilename, n_loc=3,
                                mode="w") as dataset:
            data = {"test": np.arange(3)}
            dataset.write(1,
                          data,
                          dates,
                          loc_descr="first station",
                          lon=1,
                          lat=1,
                          alt=1)
            dataset.write(2,
                          data,
                          dates,
                          loc_descr="second station",
                          lon=2,
                          lat=2,
                          alt=2)
            dataset.write(3,
                          data,
                          dates,
                          loc_descr="third station",
                          lon=3,
                          lat=3,
                          alt=3)

        with ContiguousRaggedTs(self.testfilename) as dataset:
            assert dataset.time_units == "days since 1900-01-01 00:00:00"
            data = dataset.read_all(1)
            nptest.assert_array_equal(data["test"], np.arange(3))
            nptest.assert_array_equal(data["time"], dates)

    def test_unlim_loc_file_writing(self):

        dates = np.array(
            [datetime(2007, 1, 1),
             datetime(2007, 2, 1),
             datetime(2007, 3, 1)])

        with ContiguousRaggedTs(self.testfilename, mode="w") as dataset:
            data = {"test": np.arange(3)}
            dataset.write(1,
                          data,
                          dates,
                          loc_descr="first station",
                          lon=1,
                          lat=1,
                          alt=1)
            dataset.write(2,
                          data,
                          dates,
                          loc_descr="second station",
                          lon=2,
                          lat=2,
                          alt=2)
            dataset.write(3,
                          data,
                          dates,
                          loc_descr="third station",
                          lon=3,
                          lat=3,
                          alt=3)

        with ContiguousRaggedTs(self.testfilename) as dataset:
            data = dataset.read_all(1)
            nptest.assert_array_equal(data["test"], np.arange(3))
            nptest.assert_array_equal(data["time"], dates)


class DatasetIndexedTest(unittest.TestCase):

    def setUp(self):
        self.testfilename = os.path.join(mkdtemp(), "test.nc")

    def tearDown(self):
        os.remove(self.testfilename)

    def test_file_writing(self):

        with IndexedRaggedTs(self.testfilename, n_loc=3, mode="w") as dataset:
            for n_data in [2, 5, 6]:
                for location in [1, 2, 3]:
                    data = {"test": np.arange(n_data)}
                    base = datetime(2007, 1, n_data)
                    dates = np.array(
                        [base + timedelta(hours=i) for i in range(n_data)])
                    dataset.write(location,
                                  data,
                                  dates,
                                  loc_descr="first station",
                                  lon=location,
                                  lat=location,
                                  alt=location)

        with IndexedRaggedTs(self.testfilename) as dataset:
            data = dataset.read_all(1)
            nptest.assert_array_equal(
                data["test"],
                np.concatenate([np.arange(2),
                                np.arange(5),
                                np.arange(6)]))
            test_dates = []
            for n_data in [2, 5, 6]:
                base = datetime(2007, 1, n_data)
                test_dates.append(
                    np.array(
                        [base + timedelta(hours=i) for i in range(n_data)]))
            dates = np.concatenate(test_dates)
            nptest.assert_array_equal(data["time"], dates)

    def test_file_writing_multiple_points_at_once(self):
        """
        Write multiple points at once. This means that we can have multiple
        locations more than once. Dates and data must be in the same order as
        locations. This mean we only need to translate from locations to index
        and can write the data as is.
        """
        with IndexedRaggedTs(self.testfilename, n_loc=3, mode="w") as dataset:
            locations = np.array([1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3])
            data = {
                "test":
                np.concatenate([np.arange(2),
                                np.arange(5),
                                np.arange(6)])
            }
            dates = []
            for n_data in [2, 5, 6]:
                base = datetime(2007, 1, n_data)
                dates.append(
                    np.array(
                        [base + timedelta(hours=i) for i in range(n_data)]))
            dates = np.concatenate(dates)
            dataset.write(locations,
                          data,
                          dates,
                          loc_descr=["first station"] * 13,
                          lon=locations,
                          lat=locations,
                          alt=locations)

        with IndexedRaggedTs(self.testfilename) as dataset:
            for gpis, n_data in zip([1, 2, 3], [2, 5, 6]):
                data = dataset.read_all(gpis)
                nptest.assert_array_equal(data["test"], np.arange(n_data))
                test_dates = []
                base = datetime(2007, 1, n_data)
                test_dates.append(
                    np.array(
                        [base + timedelta(hours=i) for i in range(n_data)]))
                dates = np.concatenate(test_dates)
                nptest.assert_array_equal(data["time"], dates)

    def test_file_writing_multiple_points_at_once_two_steps(self):
        """
        Write multiple points at once. Add more during an append.
        """
        with IndexedRaggedTs(self.testfilename, n_loc=4, mode="w") as dataset:
            locations = np.array([1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3])
            data = {
                "test":
                np.concatenate([np.arange(2),
                                np.arange(5),
                                np.arange(6)])
            }
            dates = []
            for n_data in [2, 5, 6]:
                base = datetime(2007, 1, n_data)
                dates.append(
                    np.array(
                        [base + timedelta(hours=i) for i in range(n_data)]))
            dates = np.concatenate(dates)
            dataset.write(locations,
                          data,
                          dates,
                          loc_descr=["first station"] * 13,
                          lon=locations,
                          lat=locations,
                          alt=locations)

        with IndexedRaggedTs(self.testfilename, n_loc=4, mode="a") as dataset:
            locations = np.array([1, 1, 4, 4])
            data = {"test": np.concatenate([np.arange(2), np.arange(2)])}
            dates = []
            for n_data in [2, 2]:
                base = datetime(2007, 2, n_data)
                dates.append(
                    np.array(
                        [base + timedelta(hours=i) for i in range(n_data)]))
            dates = np.concatenate(dates)
            dataset.write(locations,
                          data,
                          dates,
                          loc_descr=["first station"] * 4,
                          lon=locations,
                          lat=locations,
                          alt=locations)

        with IndexedRaggedTs(self.testfilename) as dataset:
            for gpis, n_data, base_month in zip([1, 2, 3, 4], [2, 5, 6, 2],
                                                [1, 1, 1, 2]):
                data = dataset.read_all(gpis)
                if gpis == 1:
                    nptest.assert_array_equal(
                        data["test"],
                        np.concatenate([np.arange(n_data),
                                        np.arange(n_data)]))
                else:
                    nptest.assert_array_equal(data["test"], np.arange(n_data))
                test_dates = []
                base = datetime(2007, base_month, n_data)
                test_dates.append(
                    np.array(
                        [base + timedelta(hours=i) for i in range(n_data)]))
                if gpis == 1:
                    base = datetime(2007, 2, n_data)
                    test_dates.append(
                        np.array([
                            base + timedelta(hours=i) for i in range(n_data)
                        ]))

                dates = np.concatenate(test_dates)
                nptest.assert_array_equal(data["time"], dates)

    def test_file_writing_multiple_points_at_once_two_steps_recarray_input(
            self):
        """
        Write multiple points at once. Add more during an append. Use record arrays as input.
        """
        with IndexedRaggedTs(self.testfilename, n_loc=4, mode="w") as dataset:
            locations = np.array([1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3])
            data = {
                "test":
                np.concatenate([np.arange(2),
                                np.arange(5),
                                np.arange(6)])
            }
            dates = []
            for n_data in [2, 5, 6]:
                base = datetime(2007, 1, n_data)
                dates.append(
                    np.array(
                        [base + timedelta(hours=i) for i in range(n_data)]))
            dates = np.concatenate(dates)
            data = np.array(data["test"],
                            dtype={
                                "names": ["test"],
                                "formats": ["i8"]
                            })
            dataset.write(locations,
                          data,
                          dates,
                          loc_descr=["first station"] * 13,
                          lon=locations,
                          lat=locations,
                          alt=locations)

        with IndexedRaggedTs(self.testfilename, n_loc=4, mode="a") as dataset:
            locations = np.array([1, 1, 4, 4])
            data = {"test": np.concatenate([np.arange(2), np.arange(2)])}
            dates = []
            for n_data in [2, 2]:
                base = datetime(2007, 2, n_data)
                dates.append(
                    np.array(
                        [base + timedelta(hours=i) for i in range(n_data)]))
            dates = np.concatenate(dates)

            data = np.array(data["test"],
                            dtype={
                                "names": ["test"],
                                "formats": ["i8"]
                            })
            dataset.write(locations,
                          data,
                          dates,
                          loc_descr=["first station"] * 4,
                          lon=locations,
                          lat=locations,
                          alt=locations)

        with IndexedRaggedTs(self.testfilename) as dataset:
            for gpis, n_data, base_month in zip([1, 2, 3, 4], [2, 5, 6, 2],
                                                [1, 1, 1, 2]):
                data = dataset.read_all(gpis)
                if gpis == 1:
                    nptest.assert_array_equal(
                        data["test"],
                        np.concatenate([np.arange(n_data),
                                        np.arange(n_data)]))
                else:
                    nptest.assert_array_equal(data["test"], np.arange(n_data))
                test_dates = []
                base = datetime(2007, base_month, n_data)
                test_dates.append(
                    np.array(
                        [base + timedelta(hours=i) for i in range(n_data)]))
                if gpis == 1:
                    base = datetime(2007, 2, n_data)
                    test_dates.append(
                        np.array([
                            base + timedelta(hours=i) for i in range(n_data)
                        ]))

                dates = np.concatenate(test_dates)
                nptest.assert_array_equal(data["time"], dates)

    def test_unlim_loc_file_writing(self):

        with IndexedRaggedTs(self.testfilename, mode="w") as dataset:
            for n_data in [2, 5, 6]:
                for location in [1, 2, 3]:
                    data = {"test": np.arange(n_data)}
                    base = datetime(2007, 1, n_data)
                    dates = np.array(
                        [base + timedelta(hours=i) for i in range(n_data)])
                    dataset.write(location,
                                  data,
                                  dates,
                                  loc_descr="first station",
                                  lon=location,
                                  lat=location,
                                  alt=location)

        with IndexedRaggedTs(self.testfilename) as dataset:
            data = dataset.read_all(1)
            nptest.assert_array_equal(
                data["test"],
                np.concatenate([np.arange(2),
                                np.arange(5),
                                np.arange(6)]))
            test_dates = []
            for n_data in [2, 5, 6]:
                base = datetime(2007, 1, n_data)
                test_dates.append(
                    np.array(
                        [base + timedelta(hours=i) for i in range(n_data)]))
            dates = np.concatenate(test_dates)
            nptest.assert_array_equal(data["time"], dates)

    def test_file_writing_with_attributes(self):

        with IndexedRaggedTs(self.testfilename, n_loc=3, mode="w") as dataset:
            for n_data in [2, 5, 6]:
                for location in [1, 2, 3]:

                    data = {"test": np.arange(n_data)}
                    base = datetime(2007, 1, n_data)
                    dates = np.array(
                        [base + timedelta(hours=i) for i in range(n_data)])
                    dataset.write(location,
                                  data,
                                  dates,
                                  loc_descr="first station",
                                  lon=0,
                                  lat=0,
                                  alt=5,
                                  attributes={"testattribute": "teststring"})

        with IndexedRaggedTs(self.testfilename) as dataset:
            data = dataset.read_all(1)
            assert dataset.dataset.variables[
                "test"].testattribute == "teststring"
            nptest.assert_array_equal(
                data["test"],
                np.concatenate([np.arange(2),
                                np.arange(5),
                                np.arange(6)]))

            test_dates = []
            for n_data in [2, 5, 6]:
                base = datetime(2007, 1, n_data)
                test_dates.append(
                    np.array(
                        [base + timedelta(hours=i) for i in range(n_data)]))
            dates = np.concatenate(test_dates)
            nptest.assert_array_equal(data["time"], dates)


class DatasetGriddedTsTests(unittest.TestCase):

    def setUp(self):
        self.testdatapath = os.path.join(mkdtemp())
        self.testfilename = os.path.join(self.testdatapath, "0107.nc")
        self.grid = grids.genreg_grid().to_cell_grid()


#    def tearDown(self):
#        os.remove(self.testfilename)

    def _test_writing_with_attributes(self,
                                      ioclass,
                                      autoscale=True,
                                      dtypes=None,
                                      scale_factors=None,
                                      offsets=None,
                                      automask=True):

        dates = pd.date_range(start="2007-01-01", end="2007-02-01")

        ts = pd.DataFrame(
            {
                "var1": np.arange(len(dates)),
                "var2": np.arange(len(dates))
            },
            index=dates)

        attributes = {
            "var1": {
                "testattribute": "teststring",
                "scale_factor": 0.5
            },
            "var2": {
                "testattribute2": "teststring2"
            }
        }

        dataset = ioclass(self.testdatapath,
                          IndexedRaggedTs,
                          mode="w",
                          grid=self.grid,
                          autoscale=autoscale)
        for gpi in [10, 11, 12, 18, 19, 20]:
            dataset.write_gp(gpi,
                             ts,
                             attributes=attributes,
                             fill_values={
                                 "var1": 5,
                                 "var2": 5
                             })

        dataset = ioclass(self.testdatapath,
                          IndexedRaggedTs,
                          mode="a",
                          grid=self.grid,
                          autoscale=autoscale)
        for gpi in [13, 10]:
            dataset.write_gp(gpi, ts)

        dataset = ioclass(self.testdatapath,
                          IndexedRaggedTs,
                          grid=self.grid,
                          autoscale=autoscale,
                          automask=automask,
                          dtypes=dtypes,
                          scale_factors=scale_factors,
                          offsets=offsets)

        for gpi in [11, 12, 20]:
            ts = dataset.read(gpi)
            dtype = np.int64
            if automask:
                dtype = np.float32
            ts_should = {
                "var1": np.arange(len(dates), dtype=dtype),
                "var2": np.arange(len(dates), dtype=dtype)
            }
            if automask:
                ts_should["var1"][5] = np.nan
                ts_should["var2"][5] = np.nan

            if dtypes is not None:
                for dtype_column in dtypes:
                    if dtype_column in ts.columns:
                        ts_should[dtype_column] = ts_should[
                            dtype_column].astype(dtypes[dtype_column])

            if scale_factors is not None:
                for scale_column in scale_factors:
                    if scale_column in ts.columns:
                        ts_should[scale_column] *= scale_factors[scale_column]

            if offsets is not None:
                for offset_column in offsets:
                    if offset_column in ts.columns:
                        ts_should[offset_column] += offsets[offset_column]

            nptest.assert_array_equal(ts["var1"], ts_should["var1"])
            nptest.assert_array_equal(ts["var2"], ts_should["var2"])

    def _test_writing_with_attributes_prepared_classes(
            self,
            ioclass,
            parameters=["var1", "var2"],
            read_bulk=False,
            dtypes=None,
            scale_factors=None,
            offsets=None,
            automask=True,
            autoscale=True):

        dates = pd.date_range(start="2007-01-01", end="2007-02-01")

        ts = pd.DataFrame(
            {
                "var1": np.arange(len(dates)),
                "var2": np.arange(len(dates))
            },
            index=dates)

        attributes = {
            "var1": {
                "testattribute": "teststring"
            },
            "var2": {
                "testattribute2": "teststring2"
            }
        }

        dataset = ioclass(self.testdatapath,
                          self.grid,
                          mode="w",
                          ioclass_kws={"read_bulk": read_bulk},
                          autoscale=autoscale)
        for gpi in [10, 11, 12, 14, 15, 16, 18]:
            dataset.write(gpi,
                          ts,
                          attributes=attributes,
                          fill_values={
                              "var1": 5,
                              "var2": 5
                          })

        dataset = ioclass(self.testdatapath,
                          self.grid,
                          mode="a",
                          ioclass_kws={"read_bulk": read_bulk},
                          autoscale=autoscale)
        for gpi in [13, 10]:
            dataset.write(gpi, ts)

        dataset = ioclass(self.testdatapath,
                          self.grid,
                          mode="r",
                          ioclass_kws={"read_bulk": read_bulk},
                          parameters=parameters,
                          autoscale=autoscale,
                          automask=automask,
                          dtypes=dtypes,
                          scale_factors=scale_factors,
                          offsets=offsets)

        for gpi in [11, 12, 14, 15, 18]:
            ts = dataset.read(gpi)
            dtype = np.int64
            if automask:
                dtype = np.float32
            ts_should = {
                "var1": np.arange(len(dates), dtype=dtype),
                "var2": np.arange(len(dates), dtype=dtype)
            }
            if automask:
                ts_should["var1"][5] = np.nan
                ts_should["var2"][5] = np.nan

            if dtypes is not None:
                for dtype_column in dtypes:
                    if dtype_column in ts.columns:
                        ts_should[dtype_column] = ts_should[
                            dtype_column].astype(dtypes[dtype_column])

            if scale_factors is not None:
                for scale_column in scale_factors:
                    if scale_column in ts.columns:
                        ts_should[scale_column] *= scale_factors[scale_column]

            if offsets is not None:
                for offset_column in offsets:
                    if offset_column in ts.columns:
                        ts_should[offset_column] += offsets[offset_column]

            for parameter in parameters:
                nptest.assert_array_equal(ts[parameter], ts_should[parameter])

    def test_writing_with_attributes_GriddedContigious(self):
        self._test_writing_with_attributes_prepared_classes(
            GriddedNcContiguousRaggedTs)

    def test_writing_with_attributes_GriddedIndexed(self):
        self._test_writing_with_attributes_prepared_classes(
            GriddedNcIndexedRaggedTs)

    def test_writing_with_attributes_GriddedOrthoMulti(self):
        self._test_writing_with_attributes_prepared_classes(
            GriddedNcOrthoMultiTs)

    def test_writing_with_attributes_GriddedContigious_read_bulk(self):
        self._test_writing_with_attributes_prepared_classes(
            GriddedNcContiguousRaggedTs, read_bulk=True)

    def test_writing_with_attributes_GriddedIndexed_read_bulk(self):
        self._test_writing_with_attributes_prepared_classes(
            GriddedNcIndexedRaggedTs, read_bulk=True)

    def test_writing_with_attributes_GriddedOrthoMulti_read_bulk(self):
        self._test_writing_with_attributes_prepared_classes(
            GriddedNcOrthoMultiTs, read_bulk=True)

    def test_writing_GriddedContigious_conversion(self):
        self._test_writing_with_attributes_prepared_classes(
            GriddedNcContiguousRaggedTs,
            dtypes={"var1": np.ubyte},
            offsets={"var1": 10},
            scale_factors={"var1": 2},
            automask=False)

    def test_writing_GriddedIndexed_conversion(self):
        self._test_writing_with_attributes_prepared_classes(
            GriddedNcIndexedRaggedTs,
            dtypes={"var1": np.ubyte},
            offsets={"var1": 10},
            scale_factors={"var1": 2},
            automask=False)

    def test_writing_GriddedOrthoMulti_conversion(self):
        self._test_writing_with_attributes_prepared_classes(
            GriddedNcOrthoMultiTs,
            dtypes={"var1": np.ubyte},
            offsets={"var1": 10},
            scale_factors={"var1": 2},
            automask=False)

    def test_writing_parameters_GriddedContigious_read_bulk(self):
        self._test_writing_with_attributes_prepared_classes(
            GriddedNcContiguousRaggedTs, read_bulk=True, parameters=["var1"])

    def test_writing_parameters_GriddedIndexed_read_bulk(self):
        self._test_writing_with_attributes_prepared_classes(
            GriddedNcIndexedRaggedTs, read_bulk=True, parameters=["var1"])

    def test_writing_parameters_GriddedOrthoMulti_read_bulk(self):
        self._test_writing_with_attributes_prepared_classes(
            GriddedNcOrthoMultiTs, read_bulk=True, parameters=["var1"])

    def test_writing_parameters_GriddedContigious_conversion(self):
        self._test_writing_with_attributes_prepared_classes(
            GriddedNcContiguousRaggedTs,
            dtypes={"var1": np.ubyte},
            offsets={"var1": 10},
            scale_factors={"var1": 2},
            autoscale=False,
            parameters=["var1"],
            automask=False)

    def test_writing_parameters_GriddedIndexed_conversion(self):
        self._test_writing_with_attributes_prepared_classes(
            GriddedNcIndexedRaggedTs,
            dtypes={"var1": np.ubyte},
            offsets={"var1": 10},
            scale_factors={"var1": 2},
            autoscale=False,
            parameters=["var1"],
            automask=False)

    def test_writing_parameters_GriddedOrthoMulti_conversion(self):
        self._test_writing_with_attributes_prepared_classes(
            GriddedNcOrthoMultiTs,
            dtypes={"var1": np.ubyte},
            offsets={"var1": 10},
            scale_factors={"var1": 2},
            autoscale=False,
            parameters=["var1"],
            automask=False)

    def test_writing_parameters_GriddedContigious_autoscale_false(self):
        self._test_writing_with_attributes_prepared_classes(
            GriddedNcContiguousRaggedTs, autoscale=False)

    def test_writing_parameters_GriddedIndexed_autoscale_false(self):
        self._test_writing_with_attributes_prepared_classes(
            GriddedNcIndexedRaggedTs, autoscale=False)

    def test_writing_parameters_GriddedOrthoMulti_autoscale_false(self):
        self._test_writing_with_attributes_prepared_classes(
            GriddedNcOrthoMultiTs, autoscale=False)


class GriddedNcTsTests(unittest.TestCase):

    def setUp(self):
        """
        Create grid and temporary location for files.
        """
        self.testdatapath = os.path.join(mkdtemp())
        self.testfilenames = [
            os.path.join(self.testdatapath, "0035.nc"),
            os.path.join(self.testdatapath, "0107.nc")
        ]

        self.gpis = [1, 10, 11, 12]
        reg_grid = grids.genreg_grid().to_cell_grid()
        self.grid = reg_grid.subgrid_from_gpis(self.gpis)


#    def tearDown(self):
#        """
#        Remove temporary files.
#        """
#        for filename in self.testfilenames:
#            if os.path.isfile(filename):
#                os.remove(filename)

    def test_n_loc(self):
        """
        Test whether the number of location is correctly reset if the cell
        needs to be changed within a loop of grid points.
        """
        dates = pd.date_range(start="2007-01-01", end="2007-02-01")

        ts = pd.DataFrame(
            {
                "var1": np.arange(len(dates)),
                "var2": np.arange(len(dates))
            },
            index=dates)

        dataset = GriddedNcContiguousRaggedTs(self.testdatapath,
                                              self.grid,
                                              mode="w")
        fill_values = {"var1": 5, "var2": 5}

        for gpi in self.gpis:
            dataset.write(gpi, ts, fill_values=fill_values)

    def test_missing_file(self):
        """
        Test missing files.
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            dataset = GriddedNcContiguousRaggedTs(mkdtemp(),
                                                  self.grid,
                                                  mode="r")

            for gpi in self.gpis:
                ts = dataset.read(gpi)
                assert ts is None

    def test_datatype(self):
        """
        Test if integer datatype is correctly written into NetCDF file.
        """
        dates = pd.date_range(start="2007-01-01", end="2007-02-01")

        ts = pd.DataFrame(
            {
                "var1": np.arange(len(dates), dtype=np.int8),
                "var2": np.arange(len(dates), dtype=np.int16),
                "var3": np.arange(len(dates), dtype=np.int32),
                "var4": np.arange(len(dates), dtype=np.int64)
            },
            index=dates)

        dataset_w = GriddedNcContiguousRaggedTs(self.testdatapath,
                                                self.grid,
                                                mode="w")

        for gpi in self.gpis:
            dataset_w.write(gpi, ts)

        dataset_r = GriddedNcContiguousRaggedTs(self.testdatapath,
                                                self.grid,
                                                mode="r")

        for gpi in self.gpis:
            arr = dataset_r.read(gpi)
            assert (arr["var1"].dtype == np.int8)
            assert (arr["var2"].dtype == np.int16)
            assert (arr["var3"].dtype == np.int32)
            assert (arr["var4"].dtype == np.int64)


class GriddedNcTsTestsSimpleGrid(unittest.TestCase):

    def setUp(self):
        """
        Create grid and temporary location for files.
        """
        self.testdatapath = os.path.join(mkdtemp())
        self.testfilenames = [os.path.join(self.testdatapath, "0001.nc")]

        self.gpis = [1, 10, 11, 12]
        self.lons = [0, 0, 1, 1]
        self.lats = [1, 1, 0, 0]
        self.cells = [1, 1, 1, 1]
        self.grid = grids.CellGrid(self.lons, self.lats, self.cells, self.gpis)


#    def tearDown(self):
#        """
#        Remove temporary files.
#        """
#        for filename in self.testfilenames:
#            if os.path.isfile(filename):
#                os.remove(filename)

    def test_write_read_non_existing_data(self):
        """
        Test whether None is returned if a non existing gpi is read.
        """
        dates = pd.date_range(start="2007-01-01", end="2007-02-01")

        ts = pd.DataFrame(
            {
                "var1": np.arange(len(dates)),
                "var2": np.arange(len(dates))
            },
            index=dates)

        dataset = GriddedNcContiguousRaggedTs(self.testdatapath,
                                              self.grid,
                                              mode="w")
        fill_values = {"var1": 5, "var2": 5}

        for gpi in [1, 10, 12]:
            dataset.write(gpi, ts, fill_values=fill_values)

        dataset.close()

        dataset = GriddedNcContiguousRaggedTs(self.testdatapath,
                                              self.grid,
                                              mode="r")
        data = dataset.read(1)
        nptest.assert_array_equal(data.columns.values,
                                  np.array([u"var1", u"var2"], dtype=object))
        with pytest.raises(IOError):
            data = dataset.read(11)



# Copyright (c) 2023, TU Wien, Department of Geodesy and Geoinformation.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Vienna University of Technology,
#     Department of Geodesy and Geoinformation nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import pytest
import unittest
from tempfile import mkdtemp

import numpy as np
import numpy.testing as nptest
import pygeogrids.grids as grids

from pynetcf.point_data import PointData, GriddedPointData


class PointDataReadWriteTest(unittest.TestCase):

    def setUp(self):
        """
        Define test file.
        """
        self.fn = os.path.join(mkdtemp(), "test.nc")
        self.fn_not_written = os.path.join(mkdtemp(), "test_not_written.nc")

    def tearDown(self):
        """
        Delete test file.
        """
        try:
            os.remove(self.fn)
        except OSError:
            pass

    def test_io_str(self):
        """
        Write/read test.
        """
        loc_ids = np.arange(0, 5)
        data1 = np.array([str(n) for n in np.arange(5, 10)])
        data2 = np.arange(10, 15)

        with PointData(self.fn, mode="w", n_obs=5) as nc:
            for loc_id, d1, d2 in zip(loc_ids, data1, data2):
                nc.write(loc_id, {"var1": d1, "var2": d2})

        with PointData(self.fn) as nc:
            nptest.assert_array_equal(
                nc["var1"], np.array([str(n) for n in range(5, 10)]))
            assert nc.nc["var1"].filters()["zlib"] is False
            assert nc.nc["var2"].filters()["zlib"] is True
            nptest.assert_array_equal(nc["var1"][1], np.array(['6']))
            nptest.assert_array_equal(nc["var2"][1], np.array([11]))

    def test_io(self):
        """
        Write/read test.
        """
        loc_ids = np.arange(0, 5)
        data1 = np.arange(5, 10)
        data2 = np.arange(10, 15)

        with PointData(self.fn, mode="w", n_obs=5) as nc:
            for loc_id, d1, d2 in zip(loc_ids, data1, data2):
                nc.write(loc_id, {"var1": np.array(d1), "var2": np.array(d2)})

        with PointData(self.fn) as nc:
            nptest.assert_array_equal(nc["var1"], range(5, 10))
            assert nc.nc["var1"].filters()["zlib"] is True
            nptest.assert_array_equal(nc["var2"][1], np.array([11]))

    def test_ioerror(self):
        """
        Should raise IOError if the files does not exist.
        """
        with pytest.raises(IOError):
            PointData(self.fn_not_written)


class PointDataAppendTest(unittest.TestCase):

    def setUp(self):
        """
        Define test file.
        """
        self.fn = os.path.join(mkdtemp(), "test.nc")

    def tearDown(self):
        """
        Delete test file.
        """
        os.remove(self.fn)

    def test_append(self):
        """
        Test appending data.
        """
        with PointData(self.fn, mode="w", n_obs=10) as nc:
            for loc_id, data in zip(range(5), range(5, 10)):
                data = np.array(data)
                if loc_id == 1:
                    nc.write(loc_id, {"var1": data, "var2": data})
                elif loc_id == 4:
                    nc.write(loc_id, {"var1": data, "var3": data})
                else:
                    nc.write(loc_id, {"var1": data})

        with PointData(self.fn, mode="a") as nc:
            for loc_id, data in zip(range(5), range(5, 10)):
                data = np.array(data)
                if loc_id == 1:
                    nc.write(loc_id, {"var1": data, "var2": data})
                elif loc_id == 4:
                    nc.write(loc_id, {"var1": data, "var3": data})
                else:
                    nc.write(loc_id, {"var1": data})

        with PointData(self.fn) as nc:
            x = np.tile(range(5, 10), 2)
            nptest.assert_array_equal(nc["var1"], x)
            nptest.assert_array_equal(nc.read(1)["var2"], np.array([6, 6]))
            nptest.assert_array_equal(nc.read(4)["var3"], np.array([9, 9]))


class PointDataAppendUnlimTest(unittest.TestCase):

    def setUp(self):
        """
        Define test file.
        """
        self.fn = os.path.join(mkdtemp(), "test.nc")

    def tearDown(self):
        """
        Delete test file.
        """
        os.remove(self.fn)

    def test_append_unlim(self):
        """
        Test appending to pre - existing point data file with
        unlimited observation dimension.
        """
        with PointData(self.fn, mode="a") as nc:
            for loc_id, data in zip(range(5), range(5, 10)):
                data = np.array(data)
                if loc_id == 1:
                    nc.write(loc_id, {"var1": data, "var2": data})
                elif loc_id == 4:
                    nc.write(loc_id, {"var1": data, "var3": data})
                else:
                    nc.write(loc_id, {"var1": data})

        with PointData(self.fn, mode="a") as nc:
            for loc_id, data in zip(range(5), range(5, 10)):
                data = np.array(data)
                if loc_id == 1:
                    nc.write(loc_id, {"var1": data, "var2": data})
                elif loc_id == 4:
                    nc.write(loc_id, {"var1": data, "var3": data})
                else:
                    nc.write(loc_id, {"var1": data})

        with PointData(self.fn) as nc:
            x = np.tile(range(5, 10), 2)
            nptest.assert_array_equal(nc["var1"], x)
            nptest.assert_array_equal(nc.read(1)["var2"], np.array([6, 6]))
            nptest.assert_array_equal(nc.read(4)["var3"], np.array([9, 9]))


class PointDataMultiDimRecarrayTest(unittest.TestCase):

    def setUp(self):
        """
        Define test file.
        """
        self.fn = os.path.join(mkdtemp(), "test.nc")

    def tearDown(self):
        """
        Delete test file.
        """
        os.remove(self.fn)

    def test_io_multi_dim_recarray(self):
        """
        Test support of multi - dimensional arrays using
        numpy.dtype.metadata field from recarray.
        """
        dim_info = {
            "dims": {
                "var": ("obs", "coef", "config"),
                "var2": ("obs", "coef", "doy")
            }
        }

        data = np.zeros(4,
                        dtype=np.dtype([("var", np.float32, (3, 13)),
                                        ("var2", np.int32, (3, 366))],
                                       metadata=dim_info))

        add_dims = {"coef": 3, "config": 13, "doy": 366}

        with PointData(self.fn, mode="w", add_dims=add_dims) as nc:
            nc.write(np.arange(4), data)

        with PointData(self.fn) as nc:
            for loc_id in range(4):
                nptest.assert_array_equal(
                    nc.read(loc_id)["var"], data["var"][0])
                nptest.assert_array_equal(
                    nc.read(loc_id)["var2"], data["var2"][0])


class PointDataMultiDimDictTest(unittest.TestCase):

    def setUp(self):
        """
        Define test file.
        """
        self.fn = os.path.join(mkdtemp(), "test.nc")

    def tearDown(self):
        """
        Delete test file.
        """
        os.remove(self.fn)

    def test_io_multi_dim_dict(self):
        """
        Test support of multi - dimensional arrays using
        numpy.dtype.metadata field from dictionary.
        """
        dim_info = {"dims": {"var1": ("obs", "coef", "config")}}
        data_var = np.zeros((4, 3, 13),
                            dtype=np.dtype(np.float32, metadata=dim_info))

        dim_info = {"dims": {"var2": ("obs", "coef", "doy")}}
        data_var2 = np.zeros((4, 3, 366),
                             dtype=np.dtype(np.int32, metadata=dim_info))

        add_dims = {"coef": 3, "config": 13, "doy": 366}

        data = {"var1": data_var, "var2": data_var2}

        with PointData(self.fn, mode="w", add_dims=add_dims) as nc:
            nc.write(np.arange(4), data)


class GriddedPointDataReadWriteTest(unittest.TestCase):

    def setUp(self):
        """
        Define test file.
        """
        self.testdatapath = os.path.join(mkdtemp())
        self.testfilename = os.path.join(self.testdatapath, "0107.nc")
        self.grid = grids.genreg_grid().to_cell_grid()

    def tearDown(self):
        """
        Delete test file.
        """
        os.remove(self.testfilename)

    def test_read_write(self):
        """
        Test writing and reading of gridded PointData.
        """
        nc = GriddedPointData(self.testdatapath,
                              mode="w",
                              grid=self.grid,
                              fn_format="{:04d}.nc")

        loc_ids = [10, 11, 12]
        dataset = [1, 2, 3]

        for loc_id, data in zip(loc_ids, dataset):
            data = np.array(data)
            nc.write(loc_id, {"var1": data})

        nc.close()

        nc = GriddedPointData(self.testdatapath,
                              grid=self.grid,
                              fn_format="{:04d}.nc")

        for i, loc_id in enumerate(loc_ids):
            data = nc.read(loc_id)
            nptest.assert_equal(data["var1"], i + 1)

        nc.close()


class GriddedPointData2PointDataTest(unittest.TestCase):

    def setUp(self):
        """
        Define test file, grid and grid points.
        """
        self.gpis = [10, 11, 12, 10000, 10001, 10002, 20000, 20001, 20002]
        self.grid = grids.genreg_grid().to_cell_grid().subgrid_from_gpis(
            self.gpis)
        self.path = mkdtemp()
        self.fn_global = os.path.join(self.path, "global.nc")

    def tearDown(self):
        """
        Delete test files.
        """
        os.remove(os.path.join(self.path, "0107.nc"))
        os.remove(os.path.join(self.path, "1464.nc"))
        os.remove(os.path.join(self.path, "2046.nc"))

    def test_read_write(self):
        """
        Test re - writing gridded PointData into single file.
        """
        loc_ids = self.gpis

        with GriddedPointData(self.path,
                              mode="w",
                              grid=self.grid,
                              fn_format="{:04d}.nc") as nc:

            for loc_id in loc_ids:
                nc.write(loc_id, {"var1": np.array(loc_id)})

        with GriddedPointData(self.path, grid=self.grid,
                              fn_format="{:04d}.nc") as nc:

            nc.to_point_data(self.fn_global)

        with PointData(self.fn_global) as nc:
            nptest.assert_equal(nc["var1"][:].sort(), loc_ids.sort())


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2023, TU Wien, Department of Geodesy and Geoinformation.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Vienna University of Technology,
#     Department of Geodesy and Geoinformation nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Base classes for reading and writing time series and images in NetCDF files
using Climate Forecast Metadata Conventions (http://cfconventions.org/).
"""

import os
import time
import numpy as np
import netCDF4
import datetime


class DatasetError(Exception):
    pass


class Dataset:
    """
    NetCDF file wrapper class that makes some things easier

    Parameters
    ----------
    filename : string
        filename of netCDF file. If already exiting then it will be opened
        as read only unless the append keyword is set. if the overwrite
        keyword is set then the file will be overwritten
    name : string, optional
        will be written as a global attribute if the file is a new file
    file_format : string, optional
        file format
    mode : string, optional
        access mode. default "r"
        "r" means read-only; no data can be modified.
        "w" means write; a new file is created, an existing file with the
            same name is deleted.
        "a" and "r+" mean append (in analogy with serial files); an existing
            file is opened for reading and writing.
        Appending s to modes w, r+ or a will enable unbuffered shared access
        to NETCDF3_CLASSIC or NETCDF3_64BIT formatted files. Unbuffered
        access may be useful even if you don"t need shared access, since it
        may be faster for programs that don"t access data sequentially.
        This option is ignored for NETCDF4 and NETCDF4_CLASSIC
        formatted files.
    zlib : boolean, optional
        Default True
        if set netCDF compression will be used
    complevel : int, optional
        Default 4
        compression level used from 1(low compression) to 9(high compression)
    autoscale : bool, optional
        If disabled data will not be automatically scaled when reading and
        writing
    automask : bool, optional
        If disabled data will not be masked during reading.
        This means Fill Values will be used instead of NaN.
    """

    def __init__(self,
                 filename,
                 name=None,
                 file_format="NETCDF4",
                 mode="r",
                 zlib=True,
                 complevel=4,
                 autoscale=True,
                 automask=True):

        self.dataset_name = name
        self.filename = filename
        self.file = None
        self.file_format = file_format
        self.buf_len = 0
        self.global_attr = {}
        self.global_attr["id"] = os.path.split(self.filename)[1]
        s = "%Y-%m-%d %H:%M:%S"
        self.global_attr["date_created"] = datetime.datetime.now().strftime(s)
        if self.dataset_name is not None:
            self.global_attr["dataset_name"] = self.dataset_name
        self.zlib = zlib
        self.complevel = complevel
        self.mode = mode
        self.autoscale = autoscale
        self.automask = automask

        if self.mode == "a" and not os.path.exists(self.filename):
            self.mode = "w"
        if self.mode == "w":
            self._create_file_dir()

        try:
            self.dataset = netCDF4.Dataset(self.filename,
                                           self.mode,
                                           format=self.file_format)
        except RuntimeError:
            raise IOError(f"File {self.filename} does not exist")

        self.dataset.set_auto_scale(self.autoscale)
        self.dataset.set_auto_mask(self.automask)

    def _create_file_dir(self):
        """
        Create directory for file to sit in.
        Avoid race condition if multiple instances are
        writing files into the same directory.
        """
        path = os.path.dirname(self.filename)
        if not os.path.exists(path):
            try:
                os.makedirs(path)
            except OSError:
                time.sleep(1)
                self._create_file_dir()

    def _set_global_attr(self):
        """
        Write global attributes to NetCDF file.
        """
        self.dataset.setncatts(self.global_attr)
        self.global_attr = {}

    def create_dim(self, name, n):
        """
        Create dimension for NetCDF file.
        if it does not yet exist

        Parameters
        ----------
        name : str
            Name of the NetCDF dimension.
        n : int
            Size of the dimension.
        """
        if name not in self.dataset.dimensions.keys():
            self.dataset.createDimension(name, size=n)

    def write_var(self,
                  name,
                  data=None,
                  dim=None,
                  attr={},
                  dtype=None,
                  zlib=None,
                  complevel=None,
                  chunksizes=None,
                  **kwargs):
        """
        Create or overwrite values in a NetCDF variable. The data will be
        written to disk once flush or close is called

        Parameters
        ----------
        name : str
            Name of the NetCDF variable.
        data : np.ndarray, optional
            Array containing the data.
            if not given then the variable will be left empty
        dim : tuple, optional
            A tuple containing the dimension names.
        attr : dict, optional
            A dictionary containing the variable attributes.
        dtype: data type, string or numpy.dtype, optional
            if not given data.dtype will be used
        zlib: boolean, optional
            explicit compression for this variable
            if not given then global attribute is used
        complevel: int, optional
            explicit compression level for this variable
            if not given then global attribute is used
        chunksizes : tuple, optional
            chunksizes can be used to manually specify the
            HDF5 chunksizes for each dimension of the variable.
        """

        fill_value = None
        if "_FillValue" in attr:
            fill_value = attr.pop("_FillValue")

        if dtype is None:
            dtype = data.dtype

        if zlib is None:
            zlib = self.zlib
        if not np.issubdtype(dtype, np.number):
            # Only numeric data can be compressed
            zlib = False
        if complevel is None:
            complevel = self.complevel

        if name in self.dataset.variables.keys():
            var = self.dataset.variables[name]
        else:
            var = self.dataset.createVariable(name,
                                              dtype,
                                              dim,
                                              fill_value=fill_value,
                                              zlib=zlib,
                                              complevel=complevel,
                                              chunksizes=chunksizes,
                                              **kwargs)

        for attr_name in attr:
            attr_value = attr[attr_name]
            var.setncattr(attr_name, attr_value)

        var.set_auto_scale(self.autoscale)
        if data is not None:
            var[:] = data

    def append_var(self, name, data, **kwargs):
        """
        append data along unlimited dimension(s) of variable

        Parameters
        ----------
        name : string
            Name of variable to append to.
        data : numpy.array
            Numpy array of correct dimension.

        Raises
        ------
        IOError
            if appending to variable without unlimited dimension
        """
        if name in self.dataset.variables.keys():
            var = self.dataset.variables[name]
            dim_unlimited = []
            key = []
            for index, dim in enumerate(var.dimensions):
                unlimited = self.dataset.dimensions[dim].isunlimited()
                dim_unlimited.append(unlimited)
                if not unlimited:
                    # if the dimension is not unlimited set the slice to :
                    key.append(slice(None, None, None))
                else:
                    # if unlimited set slice of this dimension to
                    # append meaning
                    # [var.shape[index]:]
                    key.append(slice(var.shape[index], None, None))

            dim_unlimited = np.array(dim_unlimited)
            nr_unlimited = np.where(dim_unlimited)[0].size
            key = tuple(key)
            # if there are unlimited dimensions we can do an append
            if nr_unlimited > 0:
                var[key] = data
            else:
                raise IOError(
                    "Cannot append to variable that has no unlimited dimension")
        else:
            self.write_var(name, data, **kwargs)

    def read_var(self, name):
        """
        reads variable from netCDF file

        Parameters
        ----------
        name : string
            name of the variable
        """

        if self.mode in ["r", "r+", "a"]:
            if name in self.dataset.variables.keys():
                return self.dataset.variables[name][:]

    def add_global_attr(self, name, value):
        """
        Add global attribute.

        Parameters
        ----------
        name : str
            Name.
        value : str or number
            Value.
        """
        self.global_attr[name] = value

    def flush(self):
        """
        Flush data to disk.
        """
        if self.dataset is not None:
            if self.mode in ["w", "r+", "a"]:
                self._set_global_attr()
                self.dataset.sync()

    def close(self):
        """
        Flush and close file.
        """
        if self.dataset is not None:
            self.flush()
            self.dataset.close()
            self.dataset = None

    def __enter__(self):
        """
        ContextManager enter.
        """
        return self

    def __exit__(self, value_type, value, traceback):
        """
        ContextManager exit.
        """
        self.close()

    def __del__(self):
        """
        Destructor.
        """
        if hasattr(self, "dataset"):
            self.close()



# Copyright (c) 2023, TU Wien, Department of Geodesy and Geoinformation.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Vienna University of Technology,
#     Department of Geodesy and Geoinformation nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Profiling of pynetcf.
"""

import os
import unittest
from tempfile import mkdtemp

import numpy as np
import numpy.testing as nptest

import pynetcf.base as ncbase

# number of temporary files created during profiling
NO_OF_FILES = 100
# dimension(s) of DIM_SIZExDIM_SIZE arrays written to variable
DIM_SIZE = 100


class ProfilingTest(unittest.TestCase):

    def __init__(self, methodName: str = ...):
        super().__init__(methodName)
        self.testfilename = [None] * NO_OF_FILES

    def setUp(self):

        for i in range(NO_OF_FILES):
            self.testfilename[i] = os.path.join(mkdtemp(),
                                                "test" + str(i) + ".nc")

    def tearDown(self):
        for i in range(NO_OF_FILES):
            os.remove(self.testfilename[i])

    """
    def test_write_append_read_1D(self):

        with ncbase.Dataset(self.testfilename,
                            file_format="NETCDF4", mode="w") as self.dataset:
            # create unlimited Dimension
            self.dataset.create_dim("dim", None)
            self.dataset.write_var("test", np.arange(15), dim=("dim"))

        with ncbase.Dataset(self.testfilename) as self.dataset:
            data = self.dataset.read_var("test")
            nptest.assert_array_equal(data, np.arange(15))

        with ncbase.Dataset(self.testfilename, mode="a") as self.dataset:
            self.dataset.append_var("test", np.arange(15))

        with ncbase.Dataset(self.testfilename) as self.dataset:
            data = self.dataset.read_var("test")
            nptest.assert_array_equal(
                data, np.concatenate([np.arange(15), np.arange(15)]))
    """

    def test_write_read_2D(self):

        for i in range(NO_OF_FILES):
            with ncbase.Dataset(self.testfilename[i],
                                file_format="NETCDF4",
                                mode="w") as self.dataset:
                self.dataset.create_dim("dim1", DIM_SIZE)
                self.dataset.create_dim("dim2", DIM_SIZE)
                self.dataset.write_var("test",
                                       np.arange(DIM_SIZE * DIM_SIZE).reshape(
                                           (DIM_SIZE, DIM_SIZE)),
                                       dim=("dim1", "dim2"))

            with ncbase.Dataset(self.testfilename[i]) as self.dataset:
                data = self.dataset.read_var("test")
                nptest.assert_array_equal(
                    data,
                    np.arange(DIM_SIZE * DIM_SIZE).reshape(
                        (DIM_SIZE, DIM_SIZE)))

    def test_write_append_2D(self):

        for i in range(NO_OF_FILES):
            with ncbase.Dataset(self.testfilename[i],
                                file_format="NETCDF4",
                                mode="w") as self.dataset:
                self.dataset.create_dim("dim1", DIM_SIZE)
                self.dataset.create_dim("dim2", None)
                self.dataset.write_var("test",
                                       np.arange(DIM_SIZE * DIM_SIZE).reshape(
                                           (DIM_SIZE, DIM_SIZE)),
                                       dim=("dim1", "dim2"))

            with ncbase.Dataset(self.testfilename[i]) as self.dataset:
                data = self.dataset.read_var("test")
                nptest.assert_array_equal(
                    data,
                    np.arange(DIM_SIZE * DIM_SIZE).reshape(DIM_SIZE, DIM_SIZE))

            with ncbase.Dataset(self.testfilename[i],
                                mode="a") as self.dataset:
                self.dataset.append_var(
                    "test",
                    np.arange(DIM_SIZE).reshape((DIM_SIZE, 1)))

            with ncbase.Dataset(self.testfilename[i]) as self.dataset:
                data = self.dataset.read_var("test")
                nptest.assert_array_equal(
                    data,
                    np.hstack([
                        np.arange(DIM_SIZE * DIM_SIZE).reshape(
                            (DIM_SIZE, DIM_SIZE)),
                        np.arange(DIM_SIZE).reshape((DIM_SIZE, 1))
                    ]))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2023, TU Wien, Department of Geodesy and Geoinformation.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Vienna University of Technology,
#     Department of Geodesy and Geoinformation nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Abstract class providing an interface for reading and writing time series
in NetCDF files using Climate Forecast Metadata Conventions
(http://cfconventions.org/).
"""

import os
import warnings
from abc import ABC, abstractmethod

import netCDF4
import numpy as np
import pandas as pd
from pygeobase.io_base import GriddedTsBase
from pygeogrids import CellGrid

from pynetcf.base import Dataset
from pynetcf.base import DatasetError


class DatasetTs(Dataset, ABC):
    """
    Abstract class to store common methods for NetCDF time series such as
    OrthoMulti-, ContiguousRaggedArray- and IndexedRaggedArray-representation.
    Implemented according to the NetCDF CF-conventions 1.6.

    Parameters
    ----------
    filename : string
        filename of netCDF file. If already exiting then it will be opened
        as read only unless the append keyword is set. if the overwrite
        keyword is set then the file will be overwritten
    n_loc : int, optional
        number of locations that this netCDF file contains time series for
        only required for new file
    loc_dim_name : string, optional
        name of the location dimension
    obs_dim_name : string, optional
        name of the observations dimension
    loc_ids_name : string, optional
        name of variable that has the location id"s stored
    loc_descr_name : string, optional
        name of variable that has additional location information
        stored
    time_units : string, optional
        units the time axis is given in.
        Default: "days since 1900-01-01 00:00:00"
    time_var : string, optional
        name of time variable
        Default: time
    lat_var : string, optional
        name of latitude variable
        Default: lat
    lon_var : string, optional
        name of longitude variable
        Default: lon
    alt_var : string, optional
        name of altitude variable
        Default: alt
    unlim_chunksize : int, optional
        chunksize to use along unlimited dimensions, other chunksizes
        will be calculated by the netCDF library
    read_bulk : boolean, optional
        if set to True the data of all locations is read into memory,
        and subsequent calls to "read" read from the cache and not from disk
        this makes reading complete files faster#
    read_dates : boolean, optional
        if false dates will not be read automatically but only on specific
        request useable for bulk reading because currently the netCDF
        num2date routine is very slow for big datasets
    """

    def __init__(self,
                 filename,
                 n_loc=None,
                 loc_dim_name="locations",
                 obs_dim_name="time",
                 loc_ids_name="location_id",
                 loc_descr_name="location_description",
                 time_units="days since 1900-01-01 00:00:00",
                 time_var="time",
                 lat_var="lat",
                 lon_var="lon",
                 alt_var="alt",
                 unlim_chunksize=None,
                 read_bulk=False,
                 read_dates=True,
                 **kwargs):

        super(DatasetTs, self).__init__(filename, **kwargs)

        self.n_loc = n_loc

        # dimension names
        self.obs_dim_name = obs_dim_name
        self.loc_dim_name = loc_dim_name

        # location names
        self.loc_ids_name = loc_ids_name
        self.loc_descr_name = loc_descr_name

        # time, time units and location
        self.time_var = time_var
        self.time_units = self._init_time_units(time_units)
        self.lat_var = lat_var
        self.lon_var = lon_var
        self.alt_var = alt_var

        self.unlim_chunksize = unlim_chunksize

        if unlim_chunksize is not None:
            self.unlim_chunksize = [unlim_chunksize]

        self.write_offset = None

        # variable which lists the variables that should not be
        # considered time series even if they have the correct dimension
        self.not_timeseries = [self.time_var]

        # location ids, to be read upon first reading operation
        self.loc_ids_var = None

        # date variables, for OrthogonalMulitTs it can be stored
        # since it is the same for all variables in a file
        self.constant_dates = True
        self.dates = None
        self.read_dates_auto = read_dates

        if self.mode == "r":
            self.read_bulk = read_bulk
        else:
            self.read_bulk = False

        # cache location id during reading
        self.prev_loc_id = None

        # if read bulk is activated the arrays will  be read into the
        # local variables dict if it is not activated the data will be read
        # from the netCDF variables
        if not self.read_bulk:
            self.variables = self.dataset.variables
        else:
            self.variables = {}

    @abstractmethod
    def _init_dimensions(self):
        """
        Initializes the dimensions.
        """
        pass

    @abstractmethod
    def _init_lookup(self):
        """
        Initializes variables for the lookup between locations and entries in
        the time series.
        """
        pass

    def _init_time_units(self, _default_units: str) -> str:
        try:
            time_units = self.dataset['time'].getncattr('units')
        except (IndexError, AttributeError):
            time_units = _default_units
        return time_units

    def _init_location_variables(self):
        """
        Initialize location information: longitude, latitude and altitude.
        """
        self.write_var(self.lon_var,
                       data=None,
                       dim=self.loc_dim_name,
                       attr={
                           "standard_name": "longitude",
                           "long_name": "location longitude",
                           "units": "degrees_east",
                           "valid_range": (-180.0, 180.0)
                       },
                       dtype=np.float32)

        self.write_var(self.lat_var,
                       data=None,
                       dim=self.loc_dim_name,
                       attr={
                           "standard_name": "latitude",
                           "long_name": "location latitude",
                           "units": "degrees_north",
                           "valid_range": (-90.0, 90.0)
                       },
                       dtype=np.float32)

        self.write_var(self.alt_var,
                       data=None,
                       dim=self.loc_dim_name,
                       attr={
                           "standard_name": "height",
                           "long_name": "vertical distance above the "
                           "surface",
                           "units": "m",
                           "positive": "up",
                           "axis": "Z"
                       },
                       dtype=np.float32)

    def _init_location_id_and_time(self):
        """
        Initialize the dimensions and variables that are the basis of
        the format.
        """
        # make variable that contains the location id
        self.write_var(self.loc_ids_name,
                       data=None,
                       dim=self.loc_dim_name,
                       dtype=np.int64)

        self.write_var(self.loc_descr_name,
                       data=None,
                       dim=self.loc_dim_name,
                       dtype=str)

        # initialize time variable
        self.write_var(self.time_var,
                       data=None,
                       dim=self.obs_dim_name,
                       attr={
                           "standard_name": "time",
                           "long_name": "time of measurement",
                           "units": self.time_units
                       },
                       dtype=np.float64,
                       chunksizes=self.unlim_chunksize)

    def _read_loc_ids(self, force=False):
        """
        Load location ids.
        """
        if self.loc_ids_var is None or force == True:
            loc_ids_var = self.dataset.variables[self.loc_ids_name][:]
            self.loc_ids_var = np.ma.masked_array(loc_ids_var)

    def _find_free_index_pos(self):
        """
        If the index is not yet filled completely this function
        gets the id of the first free position.

        This function depends on the masked array being used if no
        data is yet in the file.

        Returns
        -------
        idx : int
            First free index position.

        Raises
        ------
        IOError
            If no free index is found.
        """
        self._read_loc_ids()

        masked = np.where(np.atleast_1d(self.loc_ids_var.mask))[0]

        # all indexes already filled
        if len(masked) == 0:
            if self.dataset.dimensions[self.loc_dim_name].isunlimited():
                idx = self.loc_ids_var.size
            else:
                raise IOError("No free index available")
        else:
            idx = np.min(masked)

        return idx

    def _get_loc_id_index(self, loc_id):
        """
        Gets index of location id in location ids variable.

        Parameters
        ----------
        loc_id : int or numpy.ndarray
            Location id.

        Returns
        -------
        loc_id_index : int
            Location id index.
        """
        self._read_loc_ids()
        loc_id = np.atleast_1d(loc_id)
        # check if the location ids are all actually in the location id
        # variable
        in1d = np.isin(loc_id, self.loc_ids_var.data, assume_unique=True)
        if loc_id[in1d].size != loc_id.size:
            raise IOError("Location not yet defined")
        loc_ids_sorted = np.argsort(self.loc_ids_var.data)
        ypos = np.searchsorted(self.loc_ids_var[loc_ids_sorted], loc_id)
        try:
            loc_id_index = loc_ids_sorted[ypos]
            # loc_id_index = np.where(loc_id == self.loc_ids_var[:, None])[0]
        except IndexError:
            # Location not yet defined:
            raise IOError("Location not yet defined")

        if loc_id_index.size != loc_id.size:
            raise IOError(
                f"Index problem {loc_id_index.size} elements found for {loc_id.size} locations")

        if loc_id.size == 1:
            loc_id_index = loc_id_index[0]
        return loc_id_index

    def _add_location(self, loc_id, lon, lat, alt=None, loc_descr=None):
        """
        add a new location to the dataset

        Paramters
        ---------
        loc_id : int or numpy.array
            location id
        lon : float or numpy.array
            longitudes of location
        lat : float or numpy.array
            longitudes of location
        alt : float or numpy.array
            altitude of location
        loc_descr : string or numpy.array
            location description
        """

        if type(loc_id) != np.ndarray:
            loc_id = np.array([loc_id])

        if type(lon) != np.ndarray:
            lon = np.array([lon])

        # netCDF library can not handle arrays of length 1 that contain only a
        # None value
        if lon.size == 1 and lon[0] is None:
            lon = None

        if type(lat) != np.ndarray:
            lat = np.array([lat])

        # netCDF library can not handle arrays of length 1 that contain only a
        # None value
        if lat.size == 1 and lat[0] is None:
            lat = None

        if alt is not None:
            if type(alt) != np.ndarray:
                alt = np.array([alt])
            # netCDF library can not handle arrays of length 1 that contain
            # onla a None value
            if alt.size == 1 and alt[0] is None:
                alt = None

        # remove location id"s that are already in the file
        locations = np.ma.compressed(
            self.dataset.variables[self.loc_ids_name][:])

        loc_count = len(locations)
        if loc_count > 0:
            loc_ids_new = np.invert(np.isin(loc_id, locations))
            if len(np.nonzero(loc_ids_new)[0]) == 0:
                # no new locations to add
                return None
        else:
            loc_ids_new = slice(None, None, None)

        # find free index position for limited location dimension
        idx = self._find_free_index_pos()

        index = np.arange(len(loc_id[loc_ids_new])) + idx

        self.dataset.variables[self.loc_ids_name][index] = loc_id[loc_ids_new]

        if lon is not None:
            self.dataset.variables[self.lon_var][index] = lon[loc_ids_new]

        if lat is not None:
            self.dataset.variables[self.lat_var][index] = lat[loc_ids_new]

        if alt is not None:
            self.dataset.variables[self.alt_var][index] = alt[loc_ids_new]

        if loc_descr is not None:

            if type(loc_descr) != np.ndarray:
                loc_descr = np.array(loc_descr)

            if len(index) == 1:
                index = int(index[0])
                loc_ids_new = 0
                self.dataset.variables[self.loc_descr_name][index] = str(
                    loc_descr)
            else:
                self.dataset.variables[self.loc_descr_name][index] = loc_descr[
                    loc_ids_new].astype(object)

        # update location ids variable after adding location
        self._read_loc_ids(force=True)

        return idx

    def _get_all_ts_variables(self):
        """
        Gets all variable names that have the self.obs_dim_name as only
        dimension indicating that they are time series observations. This
        does not include the self.time_var variable.

        Returns
        -------
        variables : list
            List of variable names.
        """
        ts_var = []

        for variable_name in self.dataset.variables:
            if variable_name not in self.not_timeseries:
                if self.obs_dim_name in \
                        self.dataset.variables[variable_name].dimensions:
                    ts_var.append(variable_name)

        return ts_var

    @abstractmethod
    def _get_index_of_ts(self, loc_id):
        """
        Get the index of a time series.

        Parameters
        ----------
        loc_id : int
            Location id.

        Returns
        -------
        loc_id_index : int
            Location id index.
        """
        pass

    @abstractmethod
    def _get_loc_ix_from_obs_ix(self, obs_ix):
        """
        Get location index from observation index. In case of OrthoMultiTs
        all measurements are taken at the same time and therefore all
        location id"s are affected.

        Parameters
        ----------
        obs_ix : int
            Observation index.

        Returns
        -------
        loc_ix : int
            Location index.
        """
        pass

    def read_time(self, loc_id):
        """
        Read the time stamps for the given location id
        in this case the location id is irrelevant since they
        all have the same timestamps
        """
        return self.dataset.variables[self.time_var][:]

    def read_dates(self, loc_id):
        """
        Read time stamps and convert them.
        """
        self.dates = netCDF4.num2date(
            self.read_time(loc_id),
            units=self.dataset.variables[self.time_var].units,
            calendar="standard",
            only_use_cftime_datetimes=False,
            only_use_python_datetimes=True)

        return self.dates.astype("datetime64[ns]")

    def _read_var_ts(self, loc_id, var):
        """
        Read a time series of a variable at a given location id

        Parameters
        ----------
        loc_id : int
            id of location, can be a grid point id or some other id
        var : string
            name of variable to read
        """
        if self.prev_loc_id != loc_id:
            index = self._get_index_of_ts(loc_id)
            self.prev_loc_index = index
        else:
            index = self.prev_loc_index

        self.prev_loc_id = loc_id

        if self.read_bulk:
            if var not in self.variables.keys():
                self.variables[var] = self.dataset.variables[var][:]

        return self.variables[var][index]

    def read(self, variables, loc_id, dates_direct=False):
        """
        reads time series of variables

        Parameters
        ----------
        variables : list or string
        loc_id : int
            location_id
        dates_direct : boolean, optional
            if True the dates are read directly from the netCDF file
            without conversion to datetime
        """
        if type(variables) != list:
            variables = [variables]

        ts = {}
        for variable in variables:
            data = self._read_var_ts(loc_id, variable)
            ts[variable] = data

        if not dates_direct:
            # only read dates if they should be read automatically
            if self.read_dates_auto:
                # only read dates if they have not been read
                # or if they are different for each location id which is
                # the case if self.constant_dates is set to False
                if self.dates is None:
                    self.read_dates(loc_id)
                if not self.constant_dates:
                    self.read_dates(loc_id)
            ts["time"] = self.dates
        else:
            if self.read_dates_auto:
                # only read dates if they have not been read
                # or if they are different for each location id which is
                # the case if self.constant_dates is set to False
                ts["time"] = self.read_time(loc_id)

        return ts

    def read_all(self, loc_id, dates_direct=False):
        """
        read a time series of all time series variables at a given location id

        Parameters
        ----------
        loc_id : int
            id of location, can be a grid point id or some other id
        dates_direct : boolean, optional
            if True the dates are read directly from the netCDF file
            without conversion to datetime

        Returns
        -------
        time_series : dict
            keys of var and time with numpy.arrays as values
        """
        ts = self.read(self._get_all_ts_variables(),
                       loc_id,
                       dates_direct=dates_direct)

        return ts

    def extend_time(self, dates, direct=False):
        """
        Extend the time dimension and variable by the given dates

        Parameters
        ----------
        dates : numpy.array of datetime objects or floats
            Timestamps.
        direct : boolean
            if true the dates are already converted into floating
            point number of correct magnitude
        """
        if direct:
            self.append_var(self.time_var, dates)
        else:
            units = self.dataset.variables[self.time_var].units
            self.append_var(
                self.time_var,
                netCDF4.date2num(dates, units=units, calendar="standard"))

    def get_time_variable_overlap(self, dates):
        """
        Figure out if a new date array has a overlap with the already
        existing time variable.

        Return the index of the existing time variable where the new dates
        should be located.

        At the moment this only handles cases where all dates are new or none
        are new.

        Parameters
        ----------
        dates : list
            List of datetime objects

        Returns
        -------
        indexes : numpy.ndarray
            Array of indexes that overlap
        """
        timevar = self.dataset.variables[self.time_var]
        if timevar.size == 0:
            indexes = np.array([0])
        else:
            try:
                indexes = netCDF4.date2index(dates, timevar)
            except ValueError:
                indexes = np.array([timevar.size])

        return indexes

    @abstractmethod
    def write(self,
              loc_id,
              data,
              dates,
              loc_descr=None,
              lon=None,
              lat=None,
              alt=None,
              fill_values=None,
              attributes=None,
              dates_direct=False):
        """
        Write time series data, if not yet existing also add location to file
        for this data format it is assumed that in each write/append cycle
        the same amount of data is added.

        Parameters
        ----------
        loc_id : int
            Location id.
        data : dict
            Dictionary with variable names as keys and
            numpy.ndarrays as values.
        dates: numpy.ndarray
            Array of datetime objects.
        attributes : dict, optional
            Dictionary of attributes that should be added to the netCDF
            variables. can also be a dict of dicts for each variable name
            as in the data dict.
        dates_direct : boolean
            If true the dates are already converted into floating
            point number of correct magnitude.
        """
        pass

    def write_all(self,
                  loc_ids,
                  data,
                  dates,
                  loc_descrs=None,
                  lons=None,
                  lats=None,
                  alts=None,
                  fill_values=None,
                  attributes=None,
                  dates_direct=False):
        """
        Write time series data in bulk, for this the user has to provide
        a 2D array with dimensions (self.nloc, dates) that is filled with
        the time series of all grid points in the file.

        Parameters
        ----------
        loc_ids : numpy.ndarray
            location ids along the first axis of the data array
        data : dict
            dictionary with variable names as keys and 2D numpy.arrays as
            values
        dates: numpy.ndarray
            Array of datetime objects with same size as second dimension of
            data arrays.
        attributes : dict, optional
            Dictionary of attributes that should be added to the netCDF
            variables. can also be a dict of dicts for each variable name as
            in the data dict.
        dates_direct : boolean
            If true the dates are already converted into floating
            point number of correct magnitude
        """
        if self.n_loc != loc_ids.size:
            raise ValueError(
                "loc_ids is not the same number of locations in the file")
        for key in data:

            if data[key].shape[1] != dates.size:
                raise IOError(
                    f"Timestamps and dataset second dimension {key} must have the same size")

            if data[key].shape[0] != self.n_loc:
                raise IOError(
                    f"Datasets first dimension {key} must have the same size as number of locations in the file")

        # make sure zip works even if one of the parameters is not given
        if lons is None:
            lons = np.repeat(None, self.n_loc)
        if lats is None:
            lats = np.repeat(None, self.n_loc)
        if alts is None:
            alts = np.repeat(None, self.n_loc)

        # find out if attributes is a dict to be used for all variables or if
        # there is a dictionary of attributes for each variable
        unique_attr = False
        if attributes is not None:
            if sorted(data.keys()) == sorted(attributes.keys()):
                unique_attr = True

        self._add_location(loc_ids, lons, lats, alts, loc_descrs)

        for key in data:

            internal_attributes = {"name": key, "coordinates": "lat lon alt"}

            if type(fill_values) == dict:
                internal_attributes["_FillValue"] = fill_values[key]

            if attributes is not None:
                if unique_attr:
                    variable_attributes = attributes[key]
                else:
                    variable_attributes = attributes

                internal_attributes.update(variable_attributes)

            if self.unlim_chunksize is None:
                chunksizes = None
            else:
                chunksizes = [self.n_loc, self.unlim_chunksize[0]]
            self.write_var(key,
                           data=None,
                           dim=(self.loc_dim_name, self.obs_dim_name),
                           attr=internal_attributes,
                           dtype=data[key].dtype,
                           chunksizes=chunksizes)

            if self.write_offset is None:
                # current shape tells us how many elements are already
                # in the file
                self.write_offset = self.dataset.variables[key].shape[1]

            _slice = slice(self.write_offset, self.write_offset + dates.size,
                           None)
            self.dataset.variables[key][:, _slice] = data[key]

        # fill up time variable
        if dates_direct:
            self.dataset.variables[self.time_var][self.write_offset:] = dates
        else:
            units = self.dataset.variables[self.time_var].units
            self.dataset.variables[self.time_var][self.write_offset:] = \
                netCDF4.date2num(dates, units=units, calendar="standard")


class OrthoMultiTs(DatasetTs):
    """
    Implementation of the Orthogonal multidimensional array representation
    of time series according to the NetCDF CF-conventions 1.6.

    Parameters
    ----------
    filename : string
        filename of netCDF file. If already exiting then it will be opened
        as read only unless the append keyword is set. if the overwrite
        keyword is set then the file will be overwritten
    n_loc : int, optional
        number of locations that this netCDF file contains time series for
        only required for new file
    loc_dim_name : string, optional
        name of the location dimension
    obs_dim_name : string, optional
        name of the observations dimension
    loc_ids_name : string, optional
        name of variable that has the location id"s stored
    loc_descr_name : string, optional
        name of variable that has additional location information
        stored
    time_units : string, optional
        units the time axis is given in.
        Default: "days since 1900-01-01 00:00:00"
    time_var : string, optional
        name of time variable
        Default: time
    lat_var : string, optional
        name of latitude variable
        Default: lat
    lon_var : string, optional
        name of longitude variable
        Default: lon
    alt_var : string, optional
        name of altitude variable
        Default: alt
    unlim_chunksize : int, optional
        chunksize to use along unlimited dimensions, other chunksizes
        will be calculated by the netCDF library
    read_bulk : boolean, optional
        if set to True the data of all locations is read into memory,
        and subsequent calls to "read" read from the cache and not from disk
        this makes reading complete files faster#
    read_dates : boolean, optional
        if false dates will not be read automatically but only on specific
        request useable for bulk reading because currently the netCDF
        num2date routine is very slow for big datasets
    """

    def __init__(self,
                 filename,
                 n_loc=None,
                 loc_dim_name="locations",
                 obs_dim_name="time",
                 loc_ids_name="location_id",
                 loc_descr_name="location_description",
                 time_units="days since 1900-01-01 00:00:00",
                 time_var="time",
                 lat_var="lat",
                 lon_var="lon",
                 alt_var="alt",
                 unlim_chunksize=None,
                 read_bulk=False,
                 read_dates=True,
                 **kwargs):

        super(OrthoMultiTs,
              self).__init__(filename, n_loc, loc_dim_name, obs_dim_name,
                             loc_ids_name, loc_descr_name, time_units,
                             time_var, lat_var, lon_var, alt_var,
                             unlim_chunksize, read_bulk, read_dates, **kwargs)

        # initialize dimensions and index_variable
        if self.mode == "w":
            self._init_dimensions()
            self._init_lookup()
            self._init_location_variables()
            self._init_location_id_and_time()

            self.global_attr["featureType"] = "timeSeries"

    def _init_dimensions(self):
        """
        Initializes the dimensions.
        """
        if self.n_loc is None:
            raise ValueError(
                "Number of locations have to be set for new OrthoMultiTs file")

        self.create_dim(self.loc_dim_name, self.n_loc)
        self.create_dim(self.obs_dim_name, None)

    def _init_lookup(self):
        """
        Initializes variables for the lookup between locations and entries in
        the time series.
        """
        # nothing to be done for OrthoMultiTs
        pass

    def _get_index_of_ts(self, loc_id):
        """
        Get the index of a time series.

        Parameters
        ----------
        loc_id : int
            Location id.

        Returns
        -------
        loc_id_index : int
            Location id index.
        """
        try:
            loc_id_index = self._get_loc_id_index(loc_id)
        except IOError:
            raise IOError(f"Index for location id #{loc_id} not found")

        _slice = (loc_id_index, slice(None, None, None))

        return _slice

    def _get_loc_ix_from_obs_ix(self, obs_ix):
        """
        Get location index from observation index. In case of OrthoMultiTs
        all measurements are taken at the same time and therefore all
        location id"s are affected.

        Parameters
        ----------
        obs_ix : int
            Observation index.

        Returns
        -------
        loc_ix : int
            Location index.
        """
        return self.read_var(self.loc_ids_name)

    def write(self,
              loc_id,
              data,
              dates,
              loc_descr=None,
              lon=None,
              lat=None,
              alt=None,
              fill_values=None,
              attributes=None,
              dates_direct=False):
        """
        Write time series data, if not yet existing also add location to file
        for this data format it is assumed that in each write/append cycle
        the same amount of data is added.

        Parameters
        ----------
        loc_id : int
            Location id.
        data : dict
            Dictionary with variable names as keys and
            numpy.ndarrays as values.
        dates: numpy.ndarray
            Array of datetime objects.
        attributes : dict, optional
            Dictionary of attributes that should be added to the netCDF
            variables. can also be a dict of dicts for each variable name
            as in the data dict.
        dates_direct : boolean
            If true the dates are already converted into floating
            point number of correct magnitude.
        """
        try:
            idx = self._get_loc_id_index(loc_id)
        except IOError:
            _ = self._add_location(loc_id, lon, lat, alt, loc_descr)
            idx = self._get_loc_id_index(loc_id)

        # find out if attributes is a dict to be used for all variables or if
        # there is a dictionary of attributes for each variable
        unique_attr = False
        if attributes is not None:
            if sorted(data.keys()) == sorted(attributes.keys()):
                unique_attr = True

        for key in data:
            if data[key].size != dates.size:
                raise IOError(
                    f"Timestamps and dataset {key} must have the same size")

        overlap_indexes = self.get_time_variable_overlap(dates)
        if len(dates) != len(overlap_indexes):
            self.extend_time(dates, direct=dates_direct)
            self.length_before_extend = overlap_indexes[-1]
        else:
            self.length_before_extend = 0

        for key in data:

            internal_attributes = {"name": key, "coordinates": "lat lon alt"}

            if type(fill_values) == dict:
                internal_attributes["_FillValue"] = fill_values[key]

            if attributes is not None:
                if unique_attr:
                    variable_attributes = attributes[key]
                else:
                    variable_attributes = attributes

                internal_attributes.update(variable_attributes)

            if self.unlim_chunksize is None:
                chunksizes = None
            else:
                chunksizes = [self.n_loc, self.unlim_chunksize[0]]
            self.write_var(key,
                           data=None,
                           dim=(self.loc_dim_name, self.obs_dim_name),
                           attr=internal_attributes,
                           dtype=data[key].dtype,
                           chunksizes=chunksizes)

            if self.write_offset is None:
                # find start of elements that are not yet filled with values
                _slice_new = slice(self.length_before_extend, None, None)
                masked = \
                    np.where(
                        self.dataset.variables[key][idx, _slice_new].mask)[0]
                # all indexes already filled
                if len(masked) == 0:
                    raise IOError("No free data slots available")
                else:
                    self.write_offset = np.min(
                        masked) + self.length_before_extend

            _slice = slice(self.write_offset, None, None)
            # has to be reshaped to 2 dimensions because it is written
            # into 2d variable otherwise netCDF library gets confused,
            # might be a bug in netCDF?
            self.dataset.variables[key][idx, _slice] = \
                data[key].reshape(1, data[key].size)


class ContiguousRaggedTs(DatasetTs):
    """
    Class that represents a Contiguous ragged array representation of
    time series according to NetCDF CF-conventions 1.6.

    Parameters
    ----------
    filename : string
        filename of netCDF file. If already exiting then it will be opened
        as read only unless the append keyword is set. if the overwrite
        keyword is set then the file will be overwritten
    n_loc : int, optional
        number of locations that this netCDF file contains time series for
        only required for new file
    n_obs : int, optional
        how many observations will be saved into this netCDF file in total
        only required for new file
    obs_loc_lut : string, optional
        variable name in the netCDF file that contains the lookup between
        observations and locations
    loc_dim_name : string, optional
        name of the location dimension
    obs_dim_name : string, optional
        name of the observations dimension
    loc_ids_name : string, optional
        name of variable that has the location id"s stored
    loc_descr_name : string, optional
        name of variable that has additional location information
        stored
    time_units : string, optional
        units the time axis is given in.
        Default: "days since 1900-01-01 00:00:00"
    time_var : string, optional
        name of time variable
        Default: time
    lat_var : string, optional
        name of latitude variable
        Default: lat
    lon_var : string, optional
        name of longitude variable
        Default: lon
    alt_var : string, optional
        name of altitude variable
        Default: alt
    """

    def __init__(self,
                 filename,
                 n_loc=None,
                 n_obs=None,
                 obs_loc_lut="row_size",
                 obs_dim_name="obs",
                 **kwargs):

        self.n_obs = n_obs
        self.obs_loc_lut = obs_loc_lut

        super(ContiguousRaggedTs, self).__init__(filename,
                                                 n_loc=n_loc,
                                                 obs_dim_name=obs_dim_name,
                                                 **kwargs)

        if self.mode == "w":
            self._init_dimensions()
            self._init_lookup()
            self._init_location_variables()
            self._init_location_id_and_time()

            self.global_attr["featureType"] = "timeSeries"

        self.constant_dates = False

    def _init_dimensions(self):
        """
        Initializes the dimensions.
        """
        self.create_dim(self.loc_dim_name, self.n_loc)
        self.create_dim(self.obs_dim_name, self.n_obs)

    def _init_lookup(self):
        """
        Initializes variables for the lookup between locations and entries in
        the time series.
        """
        attr = {
            "long_name": "number of observations at this location",
            "sample_dimension": self.obs_dim_name
        }

        self.write_var(self.obs_loc_lut,
                       data=None,
                       dim=self.loc_dim_name,
                       dtype=np.int64,
                       attr=attr,
                       chunksizes=self.unlim_chunksize)

    def _get_index_of_ts(self, loc_id):
        """
        Get slice object for time series at location loc_id.

        Parameters
        ----------
        loc_id : int
            Location id.

        Returns
        -------
        slice_obj : slice
            Slice object with start and end of time series.

        Raises
        ------
        ValueError
            If location id could not be found.
        """
        try:
            loc_id_index = self._get_loc_id_index(loc_id)
        except IOError:
            raise IOError(
                f"Index of time series for location id #{loc_id} not found")

        if self.read_bulk and self.obs_loc_lut not in self.variables:
            self.variables[self.obs_loc_lut] = np.array(
                self.dataset.variables[self.obs_loc_lut][:])
        start = np.sum(self.variables[self.obs_loc_lut][:loc_id_index])
        end = np.sum(self.variables[self.obs_loc_lut][:loc_id_index + 1])

        return slice(start, end)

    def _get_loc_ix_from_obs_ix(self, obs_ix):
        """
        Get location index from observation index.

        Parameters
        ----------
        obs_ix : int
            Observation index.

        Returns
        -------
        loc_id_index : int
            Location id index.
        """
        bins = np.hstack((0, np.cumsum(self.variables[self.obs_loc_lut])))
        loc_id_index = np.digitize(obs_ix, bins) - 1

        return loc_id_index

    def read_time(self, loc_id):
        """
        Read the time stamps for the given location id in this case it
        works like a normal time series variable.

        Returns
        -------
        time_var : np.float64
            Time variable.
        """
        return self._read_var_ts(loc_id, self.time_var)

    def write(self,
              loc_id,
              data,
              dates,
              loc_descr=None,
              lon=None,
              lat=None,
              alt=None,
              fill_values=None,
              attributes=None,
              dates_direct=False):
        """
        Write time series data, if not yet existing also add location to file.

        Parameters
        ----------
        loc_id : int
            Location id.
        data : dict
            Dictionary with variable names as keys and
            numpy.ndarrays as values.
        dates: numpy.array
            Array of datetime objects.
        attributes : dict, optional
            Dictionary of attributes that should be added to the netCDF
            variables. can also be a dict of dicts for each variable name
            as in the data dict.
        dates_direct : boolean
            If true the dates are already converted into floating
            point number of correct magnitude.
        """
        try:
            idx = self._get_loc_id_index(loc_id)
        except IOError:
            idx = self._add_location(loc_id, lon, lat, alt, loc_descr)

        # find out if attributes is a dict to be used for all variables or if
        # there is a dictionary of attributes for each variable
        unique_attr = False
        if attributes is not None:
            if sorted(data.keys()) == sorted(attributes.keys()):
                unique_attr = True

        for key in data:
            if data[key].size != dates.size:
                raise IOError(
                    f"Timestamps and dataset {key} must have the same size")

        # add number of new elements to index_var
        self.dataset.variables[self.obs_loc_lut][idx] = dates.size

        index = self._get_index_of_ts(loc_id)
        for key in data:

            internal_attributes = {
                "name": key,
                "coordinates": "time lat lon alt"
            }

            if type(fill_values) == dict:
                internal_attributes["_FillValue"] = fill_values[key]

            if attributes is not None:
                if unique_attr:
                    variable_attributes = attributes[key]
                else:
                    variable_attributes = attributes

                internal_attributes.update(variable_attributes)

            self.write_var(key,
                           data=None,
                           dim=self.obs_dim_name,
                           attr=internal_attributes,
                           dtype=data[key].dtype,
                           chunksizes=self.unlim_chunksize)
            self.dataset.variables[key][index] = data[key]

        if dates_direct:
            self.dataset.variables[self.time_var][index] = dates
        else:
            units = self.dataset.variables[self.time_var].units
            self.dataset.variables[self.time_var][index] = \
                netCDF4.date2num(dates, units=units, calendar="standard")


class IndexedRaggedTs(DatasetTs):
    """
    Class that represents a Indexed ragged array representation of time series
    according to NetCDF CF-conventions 1.6.
    """

    def __init__(self,
                 filename,
                 n_loc=None,
                 obs_loc_lut="locationIndex",
                 **kwargs):
        self.obs_loc_lut = obs_loc_lut
        super(IndexedRaggedTs, self).__init__(filename, n_loc=n_loc, **kwargs)

        if self.mode == "w":
            self._init_dimensions()
            self._init_lookup()
            self._init_location_variables()
            self._init_location_id_and_time()

            self.global_attr["featureType"] = "timeSeries"

        self.not_timeseries.append(self.obs_loc_lut)
        self.constant_dates = False

    def _init_dimensions(self):
        """
        Initializes the dimensions.
        """
        self.create_dim(self.loc_dim_name, self.n_loc)
        self.create_dim(self.obs_dim_name, None)

    def _init_lookup(self):
        """
        Initializes variables for the lookup between locations and entries
        in the time series.
        """
        attr = {
            "long_name": "which location this observation is for",
            "instance_dimension": self.loc_dim_name
        }

        self.write_var(self.obs_loc_lut,
                       data=None,
                       dim=self.obs_dim_name,
                       dtype=np.int64,
                       attr=attr,
                       chunksizes=self.unlim_chunksize)

    def _get_index_of_ts(self, loc_id):
        """
        Parameters
        ----------
        loc_id: int
            Location index.

        Raises
        ------
        IOError
            if location id could not be found
        """
        try:
            loc_ix = self._get_loc_id_index(loc_id)
        except IOError:
            msg = f"Time series for Location #{loc_id.__str__()} not found."
            raise IOError(msg)

        if self.read_bulk and self.obs_loc_lut not in self.variables:
            self.variables[self.obs_loc_lut] = self.dataset.variables[
                self.obs_loc_lut][:]
        index = np.where(self.variables[self.obs_loc_lut] == loc_ix)[0]

        if len(index) == 0:
            msg = f"Time series for Location #{loc_id.__str__()} not found."
            raise IOError(msg)

        return index

    def _get_loc_ix_from_obs_ix(self, obs_ix):
        """
        Get location index from observation index.

        Parameters
        ----------
        obs_ix : int
            Observation index.

        Returns
        -------
        loc_ix : int
            Location index.
        """
        return self.variables[self.obs_loc_lut][obs_ix]

    def read_time(self, loc_id):
        """
        Read the time stamps for the given location id in this case it
        works like a normal time series variable.

        Returns
        -------
        time_var : np.float64
            Time variable.
        """
        return self._read_var_ts(loc_id, self.time_var)

    def write(self,
              loc_id,
              data,
              dates,
              loc_descr=None,
              lon=None,
              lat=None,
              alt=None,
              fill_values=None,
              attributes=None,
              dates_direct=False):
        """
        write time series data, if not yet existing also add location to file

        Parameters
        ----------
        loc_id : int or numpy.ndarray
            location id, if it is an array the location ids have to match the
            data in the data dictionary and in the dates array. In this way data for more than
            one point can be written into the file at once.
        data : dict or numpy.recarray
            dictionary with variable names as keys and numpy.arrays as values
        dates: numpy.array
            array of datetime objects
        attributes : dict, optional
            dictionary of attributes that should be added to the netCDF
            variables. can also be a dict of dicts for each variable name as
            in the data dict.
        dates_direct : boolean
            if true the dates are already converted into floating
            point number of correct magnitude
        """
        if type(data) == np.ndarray:
            field_names = data.dtype.names
        else:
            field_names = data.keys()

        # we always want to work with arrays
        loc_id = np.atleast_1d(loc_id)
        if len(loc_id) == 1:
            loc_id = loc_id.repeat(dates.size)

        (loc_ids_uniq, loc_ids_uniq_index,
         loc_ids_uniq_lookup) = np.unique(loc_id,
                                          return_index=True,
                                          return_inverse=True)
        lon = np.atleast_1d(lon)
        lon_uniq = lon[loc_ids_uniq_index]
        lat = np.atleast_1d(lat)
        lat_uniq = lat[loc_ids_uniq_index]
        if alt is not None:
            alt = np.atleast_1d(alt)
            alt_uniq = alt[loc_ids_uniq_index]
        else:
            alt_uniq = None
        if loc_descr is not None:
            loc_descr = np.atleast_1d(loc_descr)
            loc_descr_uniq = loc_descr[loc_ids_uniq_index]
        else:
            loc_descr_uniq = None
        try:
            idx = self._get_loc_id_index(loc_id)
        except IOError:
            idx = self._add_location(loc_ids_uniq, lon_uniq, lat_uniq,
                                     alt_uniq, loc_descr_uniq)
            idx = self._get_loc_id_index(loc_id)

        # find out if attributes is a dict to be used for all variables or if
        # there is a dictionary of attributes for each variable
        unique_attr = False
        if attributes is not None:
            if sorted(data.keys()) == sorted(attributes.keys()):
                unique_attr = True

        for key in field_names:
            if data[key].size != dates.size:
                raise DatasetError(
                    f"timestamps and dataset {key} must have the same size")

        # add number of new elements to index_var
        indices = np.atleast_1d(idx)
        self.append_var(self.obs_loc_lut, indices)

        index = np.arange(len(self.variables[
            self.obs_loc_lut]))[len(self.variables[self.obs_loc_lut]) -
                                len(indices):]

        for key in field_names:

            internal_attributes = {
                "name": key,
                "coordinates": "time lat lon alt"
            }

            if type(fill_values) == dict:
                internal_attributes["_FillValue"] = fill_values[key]

            if attributes is not None:
                if unique_attr:
                    variable_attributes = attributes[key]
                else:
                    variable_attributes = attributes

                internal_attributes.update(variable_attributes)

            # does nothing if variable exists already
            self.write_var(key,
                           data=None,
                           dim=self.obs_dim_name,
                           attr=internal_attributes,
                           dtype=data[key].dtype,
                           chunksizes=self.unlim_chunksize)

            self.dataset.variables[key][index] = data[key]

        if dates_direct:
            self.dataset.variables[self.time_var][index] = dates
        else:
            units = self.dataset.variables[self.time_var].units
            self.dataset.variables[self.time_var][index] = \
                netCDF4.date2num(dates, units=units, calendar="standard")


class GriddedNcTs(GriddedTsBase):

    def __init__(self, *args, **kwargs):

        self.parameters = None
        if "parameters" in kwargs:
            self.parameters = kwargs.pop("parameters")

        self.offsets = None
        if "offsets" in kwargs:
            self.offsets = kwargs.pop("offsets")

        self.scale_factors = None
        if "scale_factors" in kwargs:
            self.scale_factors = kwargs.pop("scale_factors")

        self.dtypes = None
        if "dtypes" in kwargs:
            self.dtypes = kwargs.pop("dtypes")

        self.autoscale = True
        if "autoscale" in kwargs:
            self.autoscale = kwargs.pop("autoscale")

        self.automask = True
        if "automask" in kwargs:
            self.automask = kwargs.pop("automask")

        super(GriddedNcTs, self).__init__(*args, **kwargs)

        self.ioclass_kws.update({
            "autoscale": self.autoscale,
            "automask": self.automask
        })
        self.dates = None

        if self.ioclass == OrthoMultiTs:
            self.read_dates = False
        else:
            self.read_dates = True

    def _open(self, gp):
        """
        Open file.

        Parameters
        ----------
        gp : int
            Grid point.

        Returns
        -------
        success : boolean
            Flag if opening the file was successful.
        """
        success = True
        cell = self.grid.gpi2cell(gp)
        filename = os.path.join(self.path,
                                f"{self.fn_format.format(cell)}.nc")

        if self.mode == "r":
            if (self.previous_cell != cell) or self.fid is None:
                self.close()

                try:
                    self.fid = self.ioclass(filename,
                                            mode=self.mode,
                                            **self.ioclass_kws)
                    self.previous_cell = cell
                except (IOError, RuntimeError):
                    success = False
                    self.fid = None
                    warnings.warn(f"I/O error {filename}", RuntimeWarning)

        if self.mode in ["w", "a"]:
            if (self.previous_cell != cell) or self.fid is None:
                self.close()

                try:
                    if self.mode == "w":
                        if "n_loc" not in self.ioclass_kws:
                            n_loc = self.grid.grid_points_for_cell(
                                cell)[0].size
                            self.ioclass_kws["n_loc"] = n_loc
                    self.fid = self.ioclass(filename,
                                            mode=self.mode,
                                            **self.ioclass_kws)
                    self.previous_cell = cell
                    self.ioclass_kws.pop("n_loc", None)
                except (IOError, RuntimeError):
                    success = False
                    self.fid = None
                    warnings.warn(f"I/O error {filename}", RuntimeWarning)

        return success

    def _read_gp(self, gpi, period=None, **kwargs):
        """
        Method reads data for given gpi, additional keyword arguments
        are passed to ioclass.read

        Parameters
        ----------
        gp : int
            Grid point.
        period : list
            2 element array containing datetimes [start, end]

        Returns
        -------
        ts : pandas.DataFrame
            Time series data.
        """
        if self.mode in ["w", "a"]:
            raise IOError("trying to read file is in write/append mode")

        if not self._open(gpi):
            return None

        if self.parameters is None:
            data = self.fid.read_all(gpi, **kwargs)
        else:
            data = self.fid.read(self.parameters, gpi, **kwargs)

        if (self.dates is None) or self.read_dates or \
                (self.dates.size != data['time'].size):
            if "dates_direct" in kwargs:
                self.dates = self.fid.read_time(gpi)
            else:
                self.dates = self.fid.read_dates(gpi)

        time = self.dates

        # remove time column from dataframe, only index should contain time
        try:
            data.pop("time")
        except KeyError:
            # if the time value is not found then do nothing
            pass

        ts = pd.DataFrame(data, index=time)

        if period is not None:
            ts = ts[period[0]:period[1]]

        if self.dtypes is not None:
            for dtype_column in self.dtypes:
                if dtype_column in ts.columns:
                    try:
                        ts[dtype_column] = ts[dtype_column].astype(
                            self.dtypes[dtype_column])
                    except ValueError:
                        raise ValueError(
                            "Dtype conversion did not work. Try turning off automatic masking."
                        )

        if self.scale_factors is not None:
            for scale_column in self.scale_factors:
                if scale_column in ts.columns:
                    ts[scale_column] *= self.scale_factors[scale_column]

        if self.offsets is not None:
            for offset_column in self.offsets:
                if offset_column in ts.columns:
                    ts[offset_column] += self.offsets[offset_column]

        return ts

    def _write_gp(self, gp, data, **kwargs):
        """
        Method writing data for given gpi.

        Parameters
        ----------
        gp : int
            Grid point.
        data : pandas.DataFrame
            Time series data to write. Index has to be pandas.DateTimeIndex.
        """
        if self.mode == "r":
            raise IOError("trying to write but file is in read mode")

        self._open(gp)
        lon, lat = self.grid.gpi2lonlat(gp)

        ds = data.to_dict("series")

        for key in ds:
            ds[key] = ds[key].values

        self.fid.write(gp,
                       ds,
                       data.index.to_pydatetime(),
                       lon=lon,
                       lat=lat,
                       **kwargs)


class GriddedNcOrthoMultiTs(GriddedNcTs):

    def __init__(self, *args, **kwargs):
        kwargs["ioclass"] = OrthoMultiTs
        super(GriddedNcOrthoMultiTs, self).__init__(*args, **kwargs)


class GriddedNcContiguousRaggedTs(GriddedNcTs):

    def __init__(self, *args, **kwargs):
        kwargs["ioclass"] = ContiguousRaggedTs
        super(GriddedNcContiguousRaggedTs, self).__init__(*args, **kwargs)


class GriddedNcIndexedRaggedTs(GriddedNcTs):

    def __init__(self, *args, **kwargs):
        kwargs["ioclass"] = IndexedRaggedTs
        super(GriddedNcIndexedRaggedTs, self).__init__(*args, **kwargs)

    def write_cell(self, cell, gpi, data, datefield):
        """
        Write complete data set into cell file.

        Parameters
        ----------
        cell : int
            Cell number.
        gpi : numpy.ndarray
            Location ids.
        data : dict or numpy record array
            dictionary with variable names as keys and numpy.arrays as values
        datefield: string
            field in the data dict that contains dates in correct format
        """
        if isinstance(self.grid, CellGrid) is False:
            raise TypeError("Associated grid is not of type pygeogrids.CellGrid.")

        if self.mode != "w":
            raise ValueError("File not opened in write mode.")

        tmp_cell = np.unique(self.grid.arrcell[gpi])

        if tmp_cell.size > 1 or tmp_cell != cell:
            raise ValueError("GPIs do not correspond to given cell.")

        lons = self.grid.arrlon[gpi]
        lats = self.grid.arrlat[gpi]

        filename = os.path.join(self.path,
                                f"{self.fn_format.format(cell)}.nc")

        if os.path.isfile(filename):
            mode = "a"
        else:
            mode = "w"

        if self.previous_cell != cell:
            self.close()
            self.previous_cell = cell
            if self.mode == "w":
                if "n_loc" not in self.ioclass_kws:
                    n_loc = self.grid.grid_points_for_cell(cell)[0].size
                    self.ioclass_kws["n_loc"] = n_loc
            self.fid = self.ioclass(filename, mode=mode, **self.ioclass_kws)
            self.ioclass_kws.pop("n_loc", None)

        if type(data) != dict:
            data = {key: data[key] for key in data.dtype.names}

        dates = data[datefield]
        del data[datefield]
        self.fid.write(gpi, data, dates, lon=lons, lat=lats, dates_direct=True)
        self.close()



# Copyright (c) 2023, TU Wien, Department of Geodesy and Geoinformation.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the Vienna University of Technology,
#      Department of Geodesy and Geoinformation nor the
#      names of its contributors may be used to endorse or promote products
#      derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Image class definition for NetCDF files.
"""

import numpy as np
import netCDF4
import datetime
import pandas as pd
import pygeogrids.grids as grids

from pynetcf.base import Dataset
from pynetcf.time_series import OrthoMultiTs


class ArrayStack(OrthoMultiTs):
    """
    Class for writing stacks of arrays (1D) into netCDF.
    Array stacks are basically orthogonal multidimensional
    array representation netCDF files.
    """

    def __init__(self, filename, grid=None, times=None, mode="r", name=""):
        self.grid = grid
        self.filename = filename
        self.times = times
        self.variables = []
        self.time_var = "time"
        self.time_units = "days since 1900-01-01"
        self.time_chunksize = 1
        self.lon_chunksize = 1

        if mode in ["a", "r"]:
            super(ArrayStack, self).__init__(filename,
                                             name=name,
                                             mode=mode,
                                             read_dates=False)
            self._load_grid()
            self._load_times()

        if mode == "w":
            if grid is None:
                raise IOError("grid needs to be defined")

            super(ArrayStack, self).__init__(filename,
                                             n_loc=len(self.grid.activegpis),
                                             name=name,
                                             mode=mode,
                                             read_dates=False)

            self.dataset.variables[self.lon_var][:] = self.grid.activearrlon
            self.dataset.variables[self.lat_var][:] = self.grid.activearrlat
            self.dataset.variables[self.loc_ids_name][:] = self.grid.activegpis

        self.lat_chunksize = len(self.grid.activegpis)

    def _load_grid(self):
        """
        Load grid.
        """
        lons = self.dataset.variables[self.lon_var][:]
        lats = self.dataset.variables[self.lat_var][:]
        self.grid = grids.BasicGrid(lons, lats)

    def _load_times(self):
        """
        Load time stamps.
        """
        self.times = netCDF4.num2date(self.dataset.variables["time"][:],
                                      self.time_units)

    def write(self, gpi, data):
        """
        Write a time series into the imagestack for gpi.

        Parameters
        ----------
        self : type
            description
        gpi : int or numpy.array
            grid point indices to write to
        data : dictionary
            dictionary of int or numpy.array for each variable
            that should be written
            shape must be (len(gpi), len(times))
        """
        gpi = np.atleast_1d(gpi)

        for i, gp in enumerate(gpi):
            for var in data:
                super(ArrayStack, self).write(
                    gp, {var: np.atleast_1d(np.atleast_2d(data[var])[i, :])},
                    np.array(self.times))

    def __setitem__(self, gpi, data):
        """
        Write a time series into the imagestack for gpi.

        Parameters
        ----------
        self : type
            description
        gpi : int or numpy.array
            grid point indices to write to
        data : dictionary
            dictionary of int or numpy.array for each variable
            that should be written
            shape must be (len(gpi), len(times))
        """
        self.write(gpi, data)

    def __getitem__(self, key):
        """
        Get data for given gpi/timestamp.

        Parameters
        ----------
        key : datetime or int
            Timestamp or gpi.

        Returns
        -------
        data : pandas.DataFrame
            Data.
        """
        if type(key) == datetime.datetime:
            index = netCDF4.date2index(key,
                                       self.dataset.variables[self.time_var])
            data = {}
            for var in self._get_all_ts_variables():
                data[var] = self.dataset.variables[var][:, index]
            return data
        else:
            gpi = np.atleast_1d(key)
            for i, gp in enumerate(gpi):
                data = super(ArrayStack, self).read_all(gp)

            return pd.DataFrame(data, index=self.times)


class ImageStack(Dataset):
    """
    Class for writing stacks of 2D images into NetCDF.
    """

    def __init__(self, filename, grid=None, times=None, mode="r", name=""):
        self.grid = grid
        self.filename = filename
        self.times = times
        self.variables = []
        self.time_var = "time"
        self.time_units = "days since 1900-01-01"
        self.time_chunksize = 1
        self.lon_chunksize = 1
        self.lat_chunksize = self.grid.lat2d.shape[1]
        super(ImageStack, self).__init__(filename, name=name, mode=mode)

        if self.mode == "w":
            self._init_dimensions()
            self._init_time()
            self._init_location_variables()
        elif self.mode in ["a", "r"]:
            self._load_grid()
            self._load_variables()

    def _init_dimensions(self):
        """
        Initialize dimensions.
        """
        self.create_dim("lon", self.grid.lon2d.shape[0])
        self.create_dim("lat", self.grid.lat2d.shape[1])
        self.create_dim("time", len(self.times))

    def _load_grid(self):
        """
        Load grid.
        """
        lons = self.dataset.variables["lon"][:]
        lats = self.dataset.variables["lat"][:]
        self.grid = grids.gridfromdims(lons, lats)

    def _load_variables(self):
        """
        Load variables.
        """
        for var in self.dataset.variables:
            if self.dataset.variables[var].dimensions == ("time", "lat",
                                                          "lon"):
                self.variables.append(var)

    def _load_times(self):
        """
        Load time stamps.
        """
        self.times = netCDF4.num2date(self.dataset.variables["time"][:],
                                      self.time_units)

    def _init_time(self):
        """
        Initialize the dimensions and variables that are the basis of
        the format.
        """
        time_data = netCDF4.date2num(self.times, self.time_units)
        self.write_var(self.time_var,
                       data=time_data,
                       dim="time",
                       attr={
                           "standard_name": "time",
                           "long_name": "time of measurement",
                           "units": self.time_units
                       },
                       dtype=np.double,
                       chunksizes=[self.time_chunksize])

    def _init_location_variables(self):
        """
        Write station information, longitude, latitude and altitude
        """
        self.write_var("lon",
                       data=self.grid.lon2d[:, 0],
                       dim="lon",
                       attr={
                           "standard_name": "longitude",
                           "long_name": "location longitude",
                           "units": "degrees_east",
                           "valid_range": (-180.0, 180.0)
                       },
                       dtype=np.float32)
        self.write_var("lat",
                       data=self.grid.lat2d[0, :],
                       dim="lat",
                       attr={
                           "standard_name": "latitude",
                           "long_name": "location latitude",
                           "units": "degrees_north",
                           "valid_range": (-90.0, 90.0)
                       },
                       dtype=np.float32)

    def init_variable(self, var):
        """
        Initialize variable.
        """
        self.write_var(var,
                       data=None,
                       dim=("time", "lat", "lon"),
                       dtype=np.float32,
                       attr={"_FillValue": -9999.})

    def write(self, gpi, data):
        """
        Write a time series into the imagestack for gpi.

        Parameters
        ----------
        self : type
            description
        gpi : int or numpy.array
            grid point indices to write to
        data : dictionary
            dictionary of int or numpy.array for each variable
            that should be written
            shape must be (len(gpi), len(times))
        """
        gpi = np.atleast_1d(gpi)

        for i, gp in enumerate(gpi):
            row, column = self.grid.gpi2rowcol(gp)
            for var in data:
                if var not in self.variables:
                    self.variables.append(var)
                    self.init_variable(var)
                self.dataset.variables[var][:, row, column] = np.atleast_2d(
                    data[var])[i, :]

    def __setitem__(self, gpi, data):
        """
        Write a time series into the imagestack for gpi.

        Parameters
        ----------
        self: type
            description
        gpi: int or numpy.array
            grid point indices to write to
        data: dictionary
            dictionary of int or numpy.array for each variable
            that should be written
            shape must be (len(gpi), len(times))
        """
        self.write(gpi, data)

    def __getitem__(self, key):
        """
        Get data for given gpi.

        Parameters
        ----------
        key : int
            gpi

        Returns
        -------
        data : pandas.DataFrame
            Data.
        """
        gpi = np.atleast_1d(key)
        data = {}
        for i, gp in enumerate(gpi):
            row, column = self.grid.gpi2rowcol(gp)
            for var in self.variables:
                data[var] = self.dataset.variables[var][:, row, column]

        return pd.DataFrame(data, index=self.times)

